# beetcamp, Copyright (C) 2021 Šarūnas Nejus. Licensed under the GPLv2 or later.
# Based on beets-bandcamp: Copyright (C) 2015 Ariel George: Original implementation
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

"""Adds bandcamp album search support to the autotagger."""

from __future__ import annotations

from operator import itemgetter
from typing import TYPE_CHECKING, Any, Literal

from beets import config, plugins

from beetsplug import fetchart  # type: ignore[attr-defined]

from beetcamp import DEFAULT_CONFIG, GuruMixin
from beetcamp.helpers import NEW_METADATA_PLUGIN_CLASS, cached_patternprop
from beetcamp.http import urlify
from beetcamp.search import search_bandcamp

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from beets import IncludeLazyConfig
    from beets.autotag.hooks import AlbumInfo, TrackInfo
    from beets.library import Album, Item, Library

if not NEW_METADATA_PLUGIN_CLASS:
    from beets.plugins import BeetsPlugin as MetadataSourcePlugin
else:
    from beets.metadata_plugins import MetadataSourcePlugin

JSONDict = dict[str, Any]
CandidateType = Literal["album", "track"]


class BandcampRequestsHandler(GuruMixin):
    """A class that provides an ability to make requests and handles failures."""

    BANDCAMP_URL_PAT = cached_patternprop(r"http[^ ]+/(album|track)/")

    @classmethod
    def from_bandcamp(cls, clue: str) -> bool:
        """Check if the clue is likely to be a bandcamp url.

        We could check whether 'bandcamp' is found in the url, however, we would be
        ignoring cases where the publisher uses their own domain (for example
        https://eaux.ro) which in reality points to their Bandcamp page. Historically,
        we found that regardless of the domain, the rest of the url stays the same,
        therefore '/album/' or '/track/' is what we are looking for in a valid url here.
        """
        return bool(cls.BANDCAMP_URL_PAT.match(clue))


class BandcampAlbumArt(BandcampRequestsHandler, fetchart.RemoteArtSource):
    NAME = ID = "Bandcamp"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.config = self._config

    def get(self, album: AlbumInfo, *_: Any) -> Iterable[fetchart.Candidate]:
        """Return the url for the cover from the bandcamp album page.

        This only returns cover art urls for bandcamp albums (by id).
        """
        url = album.mb_albumid
        if not self.from_bandcamp(url):
            self._info("Not fetching art for a non-bandcamp album URL")
        elif (guru := self.guru(url)) and (url := guru.cover_art_url):
            yield self._candidate(url=url)


class BandcampPlugin(BandcampRequestsHandler, MetadataSourcePlugin):
    MAX_COMMENT_LENGTH = 4047
    LABEL_URL_IN_COMMENT = cached_patternprop(r"Visit (https:[\w/.-]+\.[a-z]+)")
    data_source = "Bandcamp"
    beets_config: IncludeLazyConfig

    def __init__(self) -> None:
        super().__init__()
        self.beets_config = config
        self.config.add(DEFAULT_CONFIG.copy())

        if self.config["truncate_comments"].get():
            self.register_listener("album_imported", self.adjust_comments_field)

        if self.config["art"]:
            self.register_listener("pluginload", self.loaded)

    def adjust_comments_field(self, lib: Library, album: Album) -> None:  # noqa: ARG002
        """If the comments field is too long, store it as album flex attr.

        Keep the first 4000 characters in the item and store the full comment as
        a flexible attribute on the album.

        This is relevant for MPD users: mpc seems to crash trying to read comments
        longer than 4047 characters (mpd 0.23.15).
        """
        items = list(album.items())
        comments = (items[0].comments or "").encode()
        if len(comments) > self.MAX_COMMENT_LENGTH:
            self._info("Truncating comments for items in album {}", album)
            album.comments = comments
            album.store()
            truncated = f"{comments[: self.MAX_COMMENT_LENGTH - 3].decode('utf-8', errors='ignore')}..."  # noqa: E501
            for item in items:
                item.comments = truncated
                item.store()

    def loaded(self) -> None:
        """Add our own artsource to the fetchart plugin."""
        for plugin in plugins.find_plugins():
            if isinstance(plugin, fetchart.FetchArtPlugin):
                if isinstance(fetchart.ART_SOURCES, set):
                    fetchart.ART_SOURCES.add(BandcampAlbumArt)
                else:
                    fetchart.ART_SOURCES[self.data_source] = BandcampAlbumArt
                    fetchart.SOURCE_NAMES[BandcampAlbumArt] = self.data_source
                    fetchart.SOURCES_ALL.append(self.data_source)
                bandcamp_fetchart = BandcampAlbumArt(self._log, self.config)
                plugin.sources = [bandcamp_fetchart, *plugin.sources]
                break

    @classmethod
    def parse_label_url(cls, text: str) -> str | None:
        return m[1] if (m := cls.LABEL_URL_IN_COMMENT.search(text)) else None

    def _find_url_in_item(self, item: Item, name: str, type_: CandidateType) -> str:
        """Try to extract release URL from the library item.

        If the item has previously been imported, `mb_albumid` (or `mb_trackid`
        for singletons) contains the release url.

        As of 2022 April, Bandcamp purchases (at least in FLAC format) contain string
        *Visit {label_url}* in the `comments` field, therefore we try our luck here.

        If it is found, then the album/track name is converted into a valid url
        representation and appended to the `label_url`. This ends up being the correct
        url except when:
            * album name has been updated on Bandcamp but the file contains the old one
            * album name does not contain a single ascii alphanumeric character
              - in reality, this becomes '--{num}' in the url, where `num` depends on
              the number of previous releases that also did not have any valid
              alphanums. Therefore, we cannot make a reliable guess here.
        """
        if (url := getattr(item, f"mb_{type_}id", "")) and self.from_bandcamp(url):
            self._info("Fetching the URL attached to the first item, {}", url)
            return url

        if (label_url := self.parse_label_url(item.comments)) and (
            urlified_name := urlify(name)
        ):
            url = f"{label_url}/{type_}/{urlified_name}"
            self._info("Trying our guess {} before searching", url)
            return url
        return ""

    def candidates(
        self,
        items: Sequence[Item],
        artist: str,
        album: str,
        va_likely: bool,
        *_: Any,
        **__: Any,
    ) -> Iterable[AlbumInfo]:
        """Return a sequence of album candidates matching given artist and album."""
        item = items[0]
        if (
            items
            and album == item.album
            and artist == item.albumartist
            and (url := self._find_url_in_item(item, album, "album"))
            and (initial_guess := self.get_album_info(url))
        ):
            yield from initial_guess
            return

        if (va_likely or "various" in artist.lower()) and (
            # user is not searching for anything specific (default search)
            item.album == album and item.artist == artist
        ):
            name, artist = item.title, item.artist
            search_type = "t"
        else:
            name = album
            search_type = "a"

        search = {"artist": artist, "name": name, "search_type": search_type}

        for url in map(itemgetter("url"), self._search(**search)):
            if albums := self.get_album_info(url):
                yield from albums

    def item_candidates(
        self, item: Item, artist: str, title: str
    ) -> Iterable[TrackInfo]:
        """Return a sequence of singleton candidates matching given artist and title."""
        if (
            item
            and title == item.title
            and artist == item.artist
            and (url := self._find_url_in_item(item, title, "track"))
            and (initial_guess := self.get_track_info(url))
        ):
            yield initial_guess
            return

        search = {"artist": artist, "name": title, "search_type": "t"}
        results = map(itemgetter("url"), self._search(**search))
        yield from filter(None, map(self.get_track_info, results))

    def album_for_id(self, album_id: str) -> AlbumInfo | None:
        """Fetch an album by its bandcamp ID."""
        if not self.from_bandcamp(album_id):
            self._info("Not a bandcamp URL, skipping")
            return None

        albums = self.get_album_info(album_id)
        if not albums:
            return None

        if len(albums) > 1:
            # get the preferred media
            preferred = self.beets_config["match"]["preferred"]["media"].get()
            pref_to_idx = dict(zip(preferred, range(len(preferred))))
            albums = sorted(albums, key=lambda x: pref_to_idx.get(x.media, 100))
        return albums[0]

    def track_for_id(self, track_id: str) -> TrackInfo | None:
        """Fetch a track by its bandcamp ID."""
        if self.from_bandcamp(track_id):
            return self.get_track_info(track_id)

        self._info("Not a bandcamp URL, skipping")
        return None

    def _search(self, **kwargs: Any) -> Iterable[JSONDict]:
        """Return a list of track/album URLs of type search_type matching the query."""
        self._info("Searching releases for {} - {}", kwargs["artist"], kwargs["name"])
        results = search_bandcamp(**kwargs, get=self._get)
        return results[: self.config["search_max"].as_number()]



"""Module with a Helpers class that contains various static, independent functions."""

from __future__ import annotations

import re
from functools import cache, partial
from itertools import chain
from operator import contains
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeAlias, TypeVar

from beets import __version__ as beets_version
from packaging.version import Version

from .genres_lookup import GENRES

ordset = dict.fromkeys

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from re import Match, Pattern

BEETS_VERSION = Version(beets_version)
ALBUMTYPES_LIST_SUPPORT = BEETS_VERSION >= Version("1.6.0")
ARTIST_LIST_FIELDS_SUPPORT = BEETS_VERSION >= Version("2.0.0")
NEW_METADATA_PLUGIN_CLASS = BEETS_VERSION >= Version("2.4.0")
GENRES_LIST_FIELD = BEETS_VERSION >= Version("2.7.0")

JSONDict = dict[str, Any]
DIGI_MEDIA = "Digital Media"
USB_TYPE_ID = 5
FORMAT_TO_MEDIA = {
    "VinylFormat": "Vinyl",
    "CDFormat": "CD",
    "CassetteFormat": "Cassette",
    "DigitalFormat": DIGI_MEDIA,
    "DVDFormat": "DVD",
    "USB Flash Drive": DIGI_MEDIA,
}

J = TypeVar("J", bound=JSONDict)
T = TypeVar("T")


class cached_classproperty(Generic[T]):
    def __init__(self, getter: Callable[..., T]) -> None:
        self.getter = getter
        self.cache: dict[type[object], T] = {}

    def __get__(self, instance: object, owner: type[object]) -> T:
        if owner not in self.cache:
            self.cache[owner] = self.getter(owner)

        return self.cache[owner]


def cached_patternprop(
    pattern: str, flags: int = 0
) -> cached_classproperty[Pattern[str]]:
    """Pattern is compiled and cached the first time it is accessed."""
    return cached_classproperty(lambda _: re.compile(pattern, flags))


class MediaInfo(NamedTuple):
    album_id: str
    name: str
    disctitle: str
    description: str

    @classmethod
    def from_format(cls, format_: JSONDict) -> MediaInfo:
        release_format = format_.get("musicReleaseFormat")

        return cls(
            format_["@id"],
            FORMAT_TO_MEDIA[release_format or "DigitalFormat"],
            "" if release_format == "DigitalFormat" else format_["name"],
            (
                ""
                if release_format == "DigitalFormat"
                else format_.get("description") or ""
            ),
        )

    @property
    def medium_count(self) -> int:
        if self.name == "Vinyl" and (
            m := re.search(
                r"[1-5](?= ?(xLP|LP|x))|single|double|triple", self.disctitle, re.I
            )
        ):
            count = m.group()
            if count.isdigit():
                return int(count)

            return {"single": 1, "double": 2, "triple": 3}[count.lower()]

        return 1


Replacement: TypeAlias = "tuple[Pattern[str], str | Callable[[Match[str]], str]]"


class Helpers:
    SPLIT_ARTISTS_PAT = cached_patternprop(r",(?= ?)| (?:[x+/-]|//|vs|and)[.]? ")
    SPLIT_ALL_ARTISTS_PAT = cached_patternprop(
        r",(?= ?)|& | (?:[X&x+/-]|//|vs|and)[.]? "
    )
    KEYWORD_SPLIT = cached_patternprop(r"[.] | #| - ")
    KEYWORD_SUBSPLIT = cached_patternprop(r"[ -]")
    FT_PAT = cached_patternprop(
        r"""
        [ ]*                            # all preceding space
        ((?P<br>[([{])|\b)              # bracket or word boundary
        (?P<ft>
            (ft|feat|featuring|(?<=\()with|w/(?![ ]you))[. ]+ # any ft variation
            (?P<ft_artist>.+?)
            (?<!mix)                    # does not end with "mix"
            (\b|['"])                   # ends with a word boundary or quote
        )
        (?(br)                          # if started with a bracket
              [])}]                     # must end with a closing bracket
            | (?=\ -\ |\ *[][)(/]|$)    # otherwise ends with of these combinations
        )
        """,
        re.I | re.VERBOSE,
    )

    @staticmethod
    @cache
    def get_replacements() -> list[Replacement]:
        rm_strings = [
            "limited edition",
            r"^EP -",
            r"\((digital )?album\)",
            r"\(single\)",
            r"\Wvinyl\W|vinyl-only|vinyl[^ ]*cd",
            "(compiled|curated) by.*",
            "compilation: ",
            r"[\[(](presented|selected) by.*",
            r"[ |-]*free download(?! \w)",
            r"[([][^])]*free\b(?!.*mix)[^])]*[])]",
            r"[([][^])]*preview[])]",
            "Various -",
            r"CD ?\d+",
            "Name Your Price:",
            "just out!",
            "- album",
            r"^\[?premiere(\]| :)",
        ]

        camelcase = re.compile(r"(?<=[a-z])(?=[A-Z])")

        def split_artist_title(m: Match[str]) -> str:
            """See for yourself.

            https://examine-archive.bandcamp.com/album/va-examine-archive-international-sampler-xmn01
            """
            artist, title = m.groups()
            artist = camelcase.sub(" ", artist)
            title = camelcase.sub(" ", title)

            return f"{artist} - {title}"

        # fmt: off
        return [
            (re.compile(rf"(([\[(])|(^| ))\*?({'|'.join(rm_strings)})(?(2)[])]|([- ]|$))", re.I), ""),  # noqa: E501
            (re.compile(r" -([^\s-])"), r" - \1"),                                          # hi -bye                   -> hi - bye  # noqa: E501
            (re.compile(r"([^\s-])- "), r"\1 - "),                                          # hi- bye                   -> hi - bye  # noqa: E501
            (re.compile(r"  +"), " "),                                                      # hi  bye                   -> hi bye  # noqa: E501
            (re.compile(r"- \("), r"("),                                                    # hi - (bye)                -> hi (bye)  # noqa: E501
            (re.compile(r"(^[^(]*\([^(]*\))\)"), r"\1"),                                    # hi (bye))                 -> (bye)  # noqa: E501
            (re.compile(r"\( +(?=\w)"), "("),                                               # ( bye)                    -> (bye)  # noqa: E501
            (re.compile(r"(?<=\w) +\)$"), ")"),                                             # (bye )                    -> (bye)  # noqa: E501
            (re.compile(r"- Reworked"), "(Reworked)"),                                      # bye - Reworked            -> bye (Reworked)  # noqa: E501
            (re.compile(r"(\([^)]+mix)$", re.I), r"\1)"),                                  # bye - (Some Mix           -> bye - (Some Mix)  # noqa: E501
            (re.compile(r'(^|- )[“"]([^”"]+)[”"]( \(|$)'), r"\1\2\3"),                      # "bye" -> bye; hi - "bye"  -> hi - bye  # noqa: E501
            (re.compile(r"\((the )?(remixes)\)", re.I), r"\2"),                             # Album (Remixes)           -> Album Remixes  # noqa: E501
            (re.compile(r"^(\[[^]-]+\]) - (([^-]|-\w)+ - ([^-]|-\w)+)$"), r"\2 \1"),        # [Remixer] - hi - bye      -> hi - bye [Remixer]  # noqa: E501
            (re.compile(r"examine-.+CD\d+_([^_-]+)[_-](.*)"), split_artist_title),          # See https://examine-archive.bandcamp.com/album/va-examine-archive-international-sampler-xmn01  # noqa: E501
            (re.compile(r'"([^"]+)" by (.+)$'), r"\2 - \1"),                                # "bye" by hi               -> hi - bye  # noqa: E501
        ]
        # fmt: on

    @classmethod
    def remove_ft(cls, text: str) -> str:
        """Remove featuring artists from the text."""
        return cls.FT_PAT.sub("", text)

    @classmethod
    def split_artists(
        cls, artists: str | Iterable[str], force: bool = False
    ) -> list[str]:
        """Split artists taking into account delimiters such as ',', '+', 'x', 'X'.

        Note: featuring artists are removed since they are not main artists.
        """
        if not isinstance(artists, str):
            artists = ", ".join(artists)

        pat = cls.SPLIT_ALL_ARTISTS_PAT if force else cls.SPLIT_ARTISTS_PAT
        split = pat.split(cls.remove_ft(artists))
        split_artists = ordset(
            a for a in map(str.strip, split) if a not in {"", "more"}
        )
        if force:
            return list(split_artists)

        for artist in list(split_artists):
            # ' & ' or ' X ' may be part of single artist name, so we need to be careful
            # here. We check whether any of the split artists appears on their own and
            # only split then
            for subartists in (s for c in "X&" if len(s := artist.split(f" {c} ")) > 1):
                if any(a in split_artists for a in subartists):
                    split_artists.pop(artist)
                    split_artists |= ordset(subartists)
        return list(split_artists)

    @classmethod
    def clean_name(cls, name: str) -> str:
        """Both album and track names are cleaned using these patterns."""
        if name:
            for pat, repl in cls.get_replacements():
                name = pat.sub(repl, name).strip()
        return name

    @classmethod
    def get_genre(
        cls, keywords: Iterable[str], config: JSONDict, label: str
    ) -> Iterable[str]:
        """Return a comma-delimited list of valid genres, using MB genres for reference.

        1. Exclude keywords that are label names, unless they are a valid MB genre

        2. Verify each keyword's (potential genre) validity w.r.t. the configured `mode`
          * classical: valid only if the _entire keyword_ matches a MB genre in the list
          * progressive: either above or if each of the words matches MB genre - since
            it is effectively a subgenre.
          * psychedelic: one of the above or if the last word is a valid MB genre.
            This allows to be flexible regarding the variety of potential genres while
            keeping away from spammy ones.

        3. Once we have the list of keywords that coming out of the mode filters,
           an additional filter is executed:
           * if a keyword is _part of another keyword_ (genre within a sub-genre),
             we keep the more specific genre, for example
             >>> get_genre(["house", "garage house", "glitch"], "classical")
             ["garage house", "glitch"]

             "garage house" is preferred over "house".
        """
        valid_mb_genre = partial(contains, GENRES)
        label_name = label.lower().replace(" ", "")
        always_include_pat = re.compile("|".join(config["always_include"]))

        def is_label_name(kw: str) -> bool:
            return kw.replace(" ", "") == label_name and not valid_mb_genre(kw)

        def is_included(kw: str) -> bool:
            return bool(always_include_pat.pattern and always_include_pat.search(kw))

        def valid_for_mode(kw: str) -> bool:
            if config["mode"] == "classical":
                return valid_mb_genre(kw)

            words = cls.KEYWORD_SUBSPLIT.split(kw)
            if config["mode"] == "progressive":
                return valid_mb_genre(kw) or all(map(valid_mb_genre, words))

            return valid_mb_genre(kw) or valid_mb_genre(words[-1])

        unique_genres: dict[str, None] = ordset([])
        # expand badly delimited keywords
        for kw in chain.from_iterable(map(cls.KEYWORD_SPLIT.split, keywords)):
            # remove full stops and hashes and ensure the expected form of 'and'
            kw_ = kw.strip("#").replace("&", "and").replace(".", "").strip()
            if not is_label_name(kw_) and (is_included(kw_) or valid_for_mode(kw_)):
                unique_genres[kw_] = None

        def within_another_genre(genre: str) -> bool:
            """Check if this genre is part of another genre.

            Remove spaces and dashes from the rest of genres and check if any of them
            contain the given genre.

            This is so that 'dark folk' is kept while 'darkfolk' is removed, and not
            the other way around.
            """
            others = ordset(g for g in unique_genres if g not in genre)
            others |= ordset(x.replace(" ", "").replace("-", "") for x in others)
            return any(genre in x for x in others)

        return (g for g in unique_genres if not within_another_genre(g))

    @staticmethod
    def unpack_props(obj: JSONDict) -> JSONDict:
        """Add all 'additionalProperty'-ies to the parent dictionary."""
        for prop in obj.get("additionalProperty") or []:
            obj[prop["name"]] = prop["value"]
        return obj

    @staticmethod
    def get_media_formats(format_list: list[JSONDict]) -> list[MediaInfo]:
        """Return filtered Bandcamp media formats as a list of MediaInfo objects.

        Formats are filtered using the following fields,
        type_id item_type  type_name          musicReleaseFormat
                a          Digital            DigitalFormat   # digital album
                b          Digital            DigitalFormat   # discography
                t          Digital            DigitalFormat   # digital track
                i          Digital            DigitalFormat   # subscription
        0       p          Other
        1       p          Compact Disc (CD)  CDFormat
        2       p          Vinyl LP           VinylFormat
        3       p          Cassette           CassetteFormat
        4       p          DVD                DVDFormat
        5       p          USB Flash Drive
        10      p          Poster/Print
        11      p          T-Shirt/Apparel
        15      p          2 x Vinyl LP       VinylFormat
        16      p          7" Vinyl           VinylFormat
        17      p          Vinyl Box Set      VinylFormat
        18      p          Other Vinyl        VinylFormat
        19      p          T-Shirt/Shirt
        20      p          Sweater/Hoodie
        """

        def valid_format(obj: JSONDict) -> bool:
            return (
                {"name", "item_type"} < set(obj)
                # not a discography
                and obj["item_type"] != "b"
                # not a subscription
                and obj["item_type"] != "i"
                # musicReleaseFormat format is given or it is a USB
                and ("musicReleaseFormat" in obj or obj["type_id"] == USB_TYPE_ID)
                # it is not a vinyl bundle
                and (obj["item_type"] != "p" or "bundle" not in obj["name"].lower())
            )

        valid_formats = filter(valid_format, map(Helpers.unpack_props, format_list))
        return list(map(MediaInfo.from_format, valid_formats))

    @staticmethod
    def check_list_fields(data: J) -> J:
        if "albumtypes" in data and not ALBUMTYPES_LIST_SUPPORT:
            data["albumtypes"] = "; ".join(data["albumtypes"])

        if not ARTIST_LIST_FIELDS_SUPPORT:
            fields = ["artists", "artists_ids", "artists_credit", "artists_sort"]
            for f in fields:
                data.pop(f)

        if not GENRES_LIST_FIELD and (genres := data.pop("genres")):
            data["genre"] = ", ".join(genres)

        if "tracks" in data:
            data["tracks"] = [Helpers.check_list_fields(t) for t in data["tracks"]]

        return data



from __future__ import annotations

import json
import logging
import webbrowser
from argparse import Action, ArgumentParser
from functools import partial
from typing import TYPE_CHECKING, Any

import httpx
from beets import config

from .helpers import cached_patternprop
from .http import http_get_text
from .metaguru import Metaguru
from .search import search_bandcamp

if TYPE_CHECKING:
    from argparse import Namespace

    from beets import IncludeLazyConfig
    from beets.autotag.hooks import AlbumInfo, TrackInfo

DEFAULT_CONFIG = {
    "include_digital_only_tracks": True,
    "search_max": 2,
    "art": False,
    "exclude_extra_fields": [],
    "genre": {
        "capitalize": False,
        "maximum": 0,
        "mode": "progressive",
        "always_include": [],
    },
    "comments_separator": "\n---\n",
    "truncate_comments": False,
}


class GuruMixin:
    ALBUM_SLUG_IN_TRACK = cached_patternprop(r'(?<=<a id="buyAlbumLink" href=")[^"]+')

    _log: logging.Logger
    config: IncludeLazyConfig

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not hasattr(self, "_log"):
            self._log = logging.getLogger(__name__)
        if not hasattr(self, "config"):
            self.config = config["beetcamp"]
        self.config.add(DEFAULT_CONFIG)

    def _exc(self, msg_template: str, *args: object) -> None:
        self._log.log(logging.WARNING, msg_template, *args, exc_info=True)

    def _info(self, msg_template: str, *args: object) -> None:
        self._log.log(logging.DEBUG, msg_template, *args, exc_info=False)

    def _get(self, url: str) -> str:
        """Return text contents of the url response."""
        try:
            return http_get_text(url)
        except httpx.HTTPError as e:
            self._info(f"HTTP error obtaining {url}: {e}")
            return ""

    def guru(self, url: str) -> Metaguru | None:
        try:
            return Metaguru.from_html(self._get(url), config=self.config.flatten())
        except (KeyError, ValueError, AttributeError, IndexError) as e:
            self._info("Failed obtaining {}: {}", url, e)
        except Exception:
            i_url = "https://github.com/snejus/beetcamp/issues/new"
            self._exc("Unexpected error obtaining {}, please report at {}", url, i_url)

        return None

    def get_album_info(self, url: str) -> list[AlbumInfo] | None:
        """Return an AlbumInfo object for a bandcamp album page.

        If track url is given by mistake, find and fetch the album url instead.
        """
        html = self._get(url)
        if html and "/track/" in url and (m := self.ALBUM_SLUG_IN_TRACK.search(html)):
            label_url = url.split(r"/track/")[0]
            url = f"{label_url}{m[0]}"

        return guru.albums if (guru := self.guru(url)) else None

    def get_track_info(self, url: str) -> TrackInfo | None:
        """Return a TrackInfo object for a bandcamp track page."""
        return guru.singleton if (guru := self.guru(url)) else None


def get_args() -> Namespace:
    parser = ArgumentParser(
        description="""Get bandcamp release metadata from the given <release-url>
or perform bandcamp search with <query>. Anything that does not start with https://
will be assumed to be a query.

Search type flags: -a for albums, -l for labels and artists, -t for tracks.
By default, all types are searched.
"""
    )

    class UrlOrQueryAction(Action):
        def __call__(
            self,
            parser: ArgumentParser,  # noqa: ARG002
            namespace: Namespace,
            values: Any,
            option_string: str | None = None,  # noqa: ARG002
        ) -> None:
            if values:
                if values.startswith("https://"):
                    target = "release_url"
                else:
                    target = "query"
                    del namespace.release_url
                setattr(namespace, target, values)

    exclusive = parser.add_mutually_exclusive_group(required=True)
    exclusive.add_argument(
        "release_url",
        action=UrlOrQueryAction,
        nargs="?",
        help="Release URL, starting with https:// OR",
    )
    exclusive.add_argument(
        "query", action=UrlOrQueryAction, default="", nargs="?", help="Search query"
    )

    store_const = partial(
        parser.add_argument, dest="search_type", action="store_const", default=""
    )
    store_const("-a", "--album", const="a", help="Search albums")
    store_const("-l", "--label", const="b", help="Search labels and artists")
    store_const("-t", "--track", const="t", help="Search tracks")
    parser.add_argument(
        "-o",
        "--open",
        action="store",
        dest="index",
        type=int,
        help="Open search result indexed by INDEX in the browser",
    )
    parser.add_argument(
        "-p",
        "--page",
        action="store",
        dest="page",
        type=int,
        default=1,
        help="The results page to show, 1 by default",
    )

    return parser.parse_args()


def main() -> None:
    args = get_args()

    search_vars = vars(args)
    index = search_vars.pop("index", None)
    if search_vars.get("query"):
        search_results = search_bandcamp(**search_vars)

        if index:
            url = search_results[index - 1]["url"]

            print(f"Opening search result number {index}: {url}")
            webbrowser.open(url)
        else:
            print(json.dumps(search_results))
    else:
        url = args.release_url
        pl = GuruMixin()
        if result := pl.get_album_info(url) or pl.get_track_info(url):
            print(json.dumps(result))
        else:
            raise AssertionError("Failed to find a release under the given url")


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import cached_property
from itertools import starmap
from typing import TYPE_CHECKING

from .helpers import cached_patternprop

if TYPE_CHECKING:
    from collections.abc import Iterable
    from re import Pattern


@dataclass
class Catalognum:
    CONSTRAINT_TEMPLATE = r"""
    (?:
      (?<![]/@.-])          # cannot be preceded by these characters
      | (?<=//)             # or preceded by at least two slashes
    )
    (?i:
      (?<!by\ )
      (?:
        \b
        (?!(?i:vol|[el]p))  # exclude anything starting with 'vol', 'ep', 'lp'
        (?-i:{})
                            # cannot end with any of the following words
        (?<!\ va\d)         # VA1
        (?<!\ va\d\d)       # VA01
        (?<!\ va\d\d\d)     # VA001
        (?<!\b20\d\d)       # a year
        \b
        (?!["'%,-])         # cannot be followed by these characters
        (?!\ [el]p)         # may not be part of an album name
        (?!\ EVER)
      )
    )
    """
    MATCH = CONSTRAINT_TEMPLATE.format(
        r"""
    (?:
          (?<![A-Z].)[A-Z]{2,}\ 0\d{2}  # MNQ 049, SOP 063, SP 040
        | [A-Z]+[. ][A-Z]\d{3,}         # M.A025, HANDS D300
        | (?:                           # ROAD6, FREELAB9
          (?<!\ )[A-Z]{4,}\d(?!\.)  # not preceded by space
          |      [A-Z]{4,}\d(?=$)   # or at the end of the line
        )
        | \d*[A-Z$]{3,}[.-]?\d{3}       # EDLX.034, HEY-101, LI$INGLE025
        | [A-Z][A-z]{2,}0\d{2}          # Fabrik038, GiBS027, PSRL_001
        | [A-Z]{3,4}(?:CD)?[.!]?\d{2,}  # TAR30, NEN.39, ZENCD30, TMF!12
        | [A-Z]{2}\d{5}                 # RM12012, DD13109
        | [A-Z]{5}\d{2}                 # PNKMN18, LBRNM11
        | [A-Z]{6,}0\d{1}               # BODYHI01, DYNMCSS01
        | [A-z]+-[A-z]+[ -]?\d{3,}(?!\ [a-z])      # AE-MJ-001.1 o-ton 119
        | [A-z]{2,3}-?0\d{2,}           # SS-023, mt001, src002
        | [A-z+]+[ ]?(?:(?i:[EL]P))\d+  # Dystopian LP01, a+w lp036
        | [a-z]+(?:cd|lp|:)\d+          # ostgutlp45, reni:7
        | [A-Z]+\d+[-_]\d{2,}           # P90-003, CC2_011
        | [A-Z]+_[A-Z]\d{1,3}           # PRL_S03
        | [A-Z]{2}(?!999)\d{3}          # ST172
        | (?:                           # SP61, MT22
          (?<!\w\ )[A-Z]{2}\d{2}        # not preceded by a word
          |        [A-Z]{2}\d{2}(?!\ \w)# or not followed by a word
        )
        | (?<=^)[A-Z]{2}\d{4}           # WS1209
        | [a-z]{2}-\d{2,}[a-z]          # sh-303d
    )
    (?: # optionally followed by
        (?:[.-]\d+)?                    # .1 in RAWVA01.1RP, -1322 in SOP 063-1322
        (?:
            (?!MIX)
            (?<=\d\d)-?[A-Z]+           # CD in IBM001CD (needs two preceding digits)
          | RP                          # RP in RAWVA01.1RP
        )?
    )?
    """
    )
    LABEL_MATCH_TEMPLATE = CONSTRAINT_TEMPLATE.format(
        r"(?i:(?<!by\ )(?<!of\ )({}[ -]?[A-Z]*\d+([A-Z]|\.\d+)*))"
    )

    # Preceded by some variation of 'Catalogue number:'."""
    header = cached_patternprop(
        r"""
        # Cat. Number: ABC123
        # (a) 'Cat'
        # (b) '. Number:'
        # (c) ' '
        # (1) 'ABC123'
        ^
        (?i:cat     # (a) starts with 'cat' (ignoring case)
          (?:       # (b) optionally match '. Number:' or similar
            (?:\W|a?l)  # punctuation or 'l' or 'al', like 'Cat ', 'Catl', 'Catal'
            .*?         # anything
          )?
        )
        \W          # (c) some sort of punctuation preceding the catalogue number
        (           # (1) catalogue number group
          [A-Z\d]{2}    # must start with two capital letters/digits
          .*?           # lazy anything
          \w            # must end with an alphanumeric char
        )
        (\W\W|$)    # match as much as possible but stop before
                    # something like ' - All right reserved'
        """,
        re.M | re.VERBOSE,
    )
    # beginning or end of line
    start_end = cached_patternprop(rf"((^{MATCH})|({MATCH}$))", re.M | re.VERBOSE)
    # enclosed by parens or square brackets, but not ending with MIX
    delimited = cached_patternprop(rf"(?:[\[(])({MATCH})(?:[])]|$)", re.VERBOSE)
    # can possibly be followed up by a second catalogue number
    anywhere = cached_patternprop(rf"({MATCH}(?:\ [-/]\ {MATCH})?)", re.VERBOSE)
    in_album_pat = cached_patternprop(
        r"""
          (^\d*[A-Z$]{{3,}}\d+)         # ^ABC123, ^A$C123
          (?:
              :\s+                      # ': '
            | \s+[|-]\s+                # ' | ', ' - '
            | -\s+                      # '- '
            | \s+-                      # ' -'
          )
          # or
        | \s[|-]\s([A-Z]{{2,}}\d+$)     # ' - ABC123$' or ' | ABC123$'
          # or
        | [([]                          # just about anything within parens or brackets
            {}
          [])]                          # closing bracket or parens
          (?!.*\[)                      # no bracket in the rest of the string
        """.format(
            CONSTRAINT_TEMPLATE.format(
                r"""
                (?!(?i:part|va|sample))
                # at least one upper letter, ends with a digit
                ([^])]*[A-Z][^])]*[0-9]+)
                """
            )
        ),
        re.VERBOSE,
    )

    label_suffix = cached_patternprop(
        r" (?:Record(?:ing)?s|Productions|Music|Official)$", re.I
    )
    punctuation = cached_patternprop(r"\W")

    release_description: str
    album: str
    label: str
    artists_and_titles: Iterable[str]

    @cached_property
    def label_variations(self) -> set[str]:
        """Return variations of the label name.

        This includes the label name without any punctuation and suffixes.
        """
        variations = {
            self.label,
            *self.label.split(" fka "),
            self.label_suffix.sub("", self.label),
        }
        variations |= {self.punctuation.sub("", v) for v in variations}

        return variations

    @cached_property
    def excluded_text(self) -> str:
        """Words that cannot be matched as a catalogue number."""
        return " ".join([*self.artists_and_titles, *self.label_variations]).lower()

    @cached_property
    def label_pattern(self) -> Pattern[str]:
        """Pattern to match catalogue numbers prefixed by the label name.

        Extend label variations with acronyms and the first word of the label.
        Return the pattern where each variation is OR'ed together.
        """
        prefixes = self.label_variations

        for prefix in [p for p in prefixes if " " in p]:
            acronym = "".join(word[0] for word in prefix.split())
            prefixes.add(acronym)

        label = self.label.strip()

        if " " in label and len(first := label.split()[0]) > 1:
            # add the first word too
            prefixes.add(first)

        str_pattern = f"(?:{'|'.join(map(re.escape, prefixes))})"

        return re.compile(self.LABEL_MATCH_TEMPLATE.format(str_pattern), re.VERBOSE)

    @cached_property
    def in_album_or_release_description(self) -> str | None:
        """Return the catalogue number found in the album name or release description.

        This is defined as a cached property so that the search is performed once for
        a certain release.
        """
        return self.find(
            [
                (self.anywhere, self.album),
                (self.label_pattern, self.album),
                (self.start_end, self.release_description),
                (self.anywhere, self.release_description),
                (self.label_pattern, self.release_description),
            ]
        )

    def search(self, pat: Pattern[str], string: str) -> str | None:
        """Search text with the given pattern and return matching catalogue number.

        This returns the first match which is not in the excluded text.
        """
        if not string:
            return None

        for m in pat.finditer(string):
            catnum = m.group(1).strip()
            if catnum.lower() not in self.excluded_text:
                return catnum

        return None

    def find(
        self, patterns_and_texts: Iterable[tuple[Pattern[str], str]]
    ) -> str | None:
        """Try getting the catalog number using supplied pattern/string pairs."""

        try:
            return next(filter(None, starmap(self.search, patterns_and_texts)))
        except StopIteration:
            return None

    def get(self, media_description: str) -> str | None:
        return (
            self.find(
                [
                    (self.header, media_description),
                    (self.header, self.release_description),
                    (self.anywhere, media_description),
                    (self.label_pattern, media_description),
                ]
            )
            or self.in_album_or_release_description
        )



"""Module with tracks parsing functionality."""

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from functools import cached_property
from os.path import commonprefix
from typing import TYPE_CHECKING, Any, ClassVar

from .helpers import cached_patternprop
from .names import Names
from .track import Track

ordset = dict.fromkeys

if TYPE_CHECKING:
    from collections.abc import Iterator


@dataclass
class Tracks:
    MIN_TRACKS_FOR_TITLE_FIX = 3
    MAX_TRACKS_WITH_ARTIST_FOR_MIS_SPLIT_FIX = 3
    MAX_TITLE_FRAGMENT_ARTISTS = 2
    DASH_INSIDE_PARENS = cached_patternprop(r"\([^)]+-[^)]+\)")
    TITLE_LIKE_ARTIST = cached_patternprop(r":|\(\d{4}\)")
    DISC_BY_LETTER: ClassVar[dict[str, int]] = {
        "A": 1,
        "B": 1,
        "C": 2,
        "D": 2,
        "E": 3,
        "F": 3,
        "G": 4,
        "H": 4,
        "I": 5,
        "J": 5,
    }
    tracklist: list[Track]
    names: Names

    def __iter__(self) -> Iterator[Track]:
        return iter(self.tracklist)

    def __len__(self) -> int:
        return len(self.tracklist)

    def __getitem__(self, idx: int) -> Track:
        return self.tracklist[idx]

    @classmethod
    def from_names(cls, names: Names) -> Tracks:
        tracks = names.json_tracks
        for track, name in zip(tracks, names.titles):
            track["name"] = name

        album_artist = names.album_artist
        if (
            len(tracks) > 1
            and names.original_album in names.common_prefix
            and album_artist != names.label
            and "," not in album_artist
        ):
            for track in tracks:
                track["album_artist"] = album_artist

        obj = cls(list(map(Track.make, tracks)), names)
        obj.handle_wild_track_alt()
        obj.fix_title_split()
        return obj

    @property
    def tracks_without_artist(self) -> list[Track]:
        return [t for t in self if not t.artist]

    @cached_property
    def first(self) -> Track:
        return self[0]

    @cached_property
    def raw_names(self) -> list[str]:
        return [j.name for j in self]

    @property
    def original_artists(self) -> list[str]:
        """Return all unique unsplit (original) main track artists."""
        return list(dict.fromkeys(j.artist for j in self))

    @property
    def artists(self) -> list[str]:
        """Return all unique split main track artists.

        "Artist1 x Artist2" -> ["Artist1", "Artist2"]
        """
        return list(dict.fromkeys(a for t in self for a in t.artists))

    @cached_property
    def lead_artists(self) -> list[str]:
        """Return all lead artists.

        Keep the first artist in collaborations:
        [A] -> [A]
        [A, A & B] -> [A]
        [A & B, A & C] -> [A]
        [A, B] -> [A, B]
        [A, B, B & C] -> [A, B]
        [A, B & C, B & D] -> [A, B]

        But keep collaborations in tact if artists do not appear on their own:
        [A & B] -> [A & B]
        [A, B & C] -> [A, B & C]
        """
        lead_artists = list(dict.fromkeys(t.lead_artist for t in self))
        unique_artists = set(self.artists)
        if not unique_artists:
            return []

        return [
            (
                a
                if a in unique_artists
                or len(collabs := [u for u in unique_artists if a in u]) > 1
                else collabs[0]
            )
            for a in lead_artists
        ]

    @property
    def collaborators(self) -> set[str]:
        """Return all unique remix and featuring artists."""
        artists = set(self.artists)
        remixers = {
            r
            for t in self
            if t.remix and (r := t.remix.artist) and all(a not in r for a in artists)
        }
        feat = {j.ft for j in self if j.ft}
        return remixers | feat

    @cached_property
    def artists_and_titles(self) -> set[str]:
        """Return a set with all artists and titles."""
        return set(self.raw_names) | self.collaborators | set(self.original_artists)

    def discard_collaborators(self, artists: list[str]) -> list[str]:
        collaborators = " ".join(self.collaborators).lower()

        return [
            a
            for a in artists
            if any(sa not in collaborators for sa in a.lower().split(" & "))
        ]

    def fix_title_split(self) -> None:
        if 1 <= len(tracks := self.tracks_without_artist) < len(self) / 2:
            for t in (t for t in tracks if t.remix):
                # reset the artist if it got removed because it's in the remix text
                if len(t.name_split) > 1:
                    t.artist = " - ".join(t.name_split[:-1])
                else:
                    t.artist = t.json_artist

        # identify a track where artist was incorrectly parsed from the title
        # while a llegitimate artist was available in the JSON data.
        not_json_artist_tracks = [
            t
            for t in self
            if t.artist
            and "," not in t.json_artist
            and not commonprefix([t.json_artist, t.artist])
            and not commonprefix([t.json_artist, t.remix.remixer if t.remix else ""])
            and t.json_artist.lower() not in t.name.lower()
        ]
        if (
            len(not_json_artist_tracks) == 1
            and len(self) >= self.MIN_TRACKS_FOR_TITLE_FIX
        ):
            for t in not_json_artist_tracks:
                t.title = f"{t.artist} - {t.title}"
                t.artist = t.json_artist

    def handle_wild_track_alt(self) -> None:
        """Handle tracks that have incorrectly parsed `track_alt` field.

        If there is a single unique `track_alt` value and multiple tracks in the
        release, assign the `track_alt` value to the `artist` or reset the `title` to
        the initial track's name.

        Clear the `track_alt` value after assignment.
        """
        unique_track_alts = {t.track_alt for t in self if t.track_alt}
        if len(unique_track_alts) == 1 and len(self) > 1:
            track_alt_tracks = [t for t in self if t.track_alt]

            same_track_alt_on_all = len(track_alt_tracks) == len(self)
            parsed_any_artists = any(t for t in self if t.artist)

            may_set_artist = parsed_any_artists or same_track_alt_on_all
            unique_track_alt = unique_track_alts.pop()
            for t in track_alt_tracks:
                if not t.artist and may_set_artist:
                    t.artist = unique_track_alt
                else:
                    t.title = t.json_item["name"]
                t.track_alt = None

    def fix_track_artists(self, albumartist: str) -> None:
        """Adjust track artists in the context of the entire album.

        Firstly, check whether the artist is 'the'. If so, prepend it to the title and
        reset the artist.

        Then, check how many tracks are missing artists. If there are 1-3 tracks
        which have it set, this is most likely because the titles had ' - ' separator
        and our logic split it into artist. For each, ensure that
            (1) Artist was not set in the JSON metadata
            (2) This string is not part of the albumartist
            (3) Albumartist is not found anywhere in the title
        If so, move this string back to the title and replace it by the albumartist.

        If only one artist is missing, check whether the title can be split by '-'
        without spaces or some other UTF-8 equivalent (most likely an alternative
        representation of a dash).

        Otherwise, use the albumartist as the default.
        """
        for t in (t for t in self if t.artist):
            # the artist cannot be 'the', so it's most likely a part of the title
            if t.artist.lower() == "the":
                t.title = f"{t.artist} {t.title}"
                t.artist = ""

        self._reset_title_fragment_artists(albumartist)

        if not self.tracks_without_artist:
            return

        if (
            1
            <= len(self) - len(self.tracks_without_artist)
            <= self.MAX_TRACKS_WITH_ARTIST_FOR_MIS_SPLIT_FIX
        ):
            aartist = albumartist.lower()
            for t in (
                t
                for t in self
                if (artist := t.artist.lower())
                and not t.json_artist
                and artist not in aartist
                and aartist not in f"{artist}{t.ft_artist.lower()}{t.title.lower()}"
            ):
                t.title = f"{t.artist} - {t.title}"
                t.artist = ""

        if (1 <= len(tracks := self.tracks_without_artist) < len(self) / 2) or (
            albumartist == "Various Artists" and len(tracks) == len(self)
        ):
            for t in tracks:
                if not self.DASH_INSIDE_PARENS.search(t.title) and (
                    len(split := t.title.split("-", 1)) > 1
                    or len(split := Names.SEPARATOR_PAT.split(t.title, 1)) > 1
                ):
                    t.artist, t.title = map(str.strip, split)

        for t in self.tracks_without_artist:
            # default to the albumartist
            t.artist = albumartist

    def _reset_title_fragment_artists(self, albumartist: str) -> None:
        """Replace parsed artists that are more likely to be title fragments.

        Notably, handle this release: https://carriez.bandcamp.com/album/satie-complete-piano-works-volume-10
        """
        # Nothing to reset against if there's no album artist
        if not albumartist:
            return

        # Collect artists that were explicitly provided in the JSON metadata
        explicit_artists = {t.json_artist.lower() for t in self if t.json_artist}
        # If tracks have explicit artists that differ from the album artist,
        # the parsed artists are likely real — not title fragments
        if explicit_artists and explicit_artists != {albumartist.lower()}:
            return

        # Only consider tracks where the artist was parsed (not from JSON metadata)
        tracks_with_parsed_artist = [t for t in self if t.artist and not t.json_artist]
        unique_artists = {t.lead_artist.lower() for t in tracks_with_parsed_artist}
        if (
            # No parsed artists to evaluate
            not unique_artists
            # Too many to be title fragments
            or len(unique_artists) > self.MAX_TITLE_FRAGMENT_ARTISTS
            # Parsed artist matches album artist - not a fragment
            or albumartist.lower() in unique_artists
            # None look like title fragments
            or not any(self.TITLE_LIKE_ARTIST.search(a) for a in unique_artists)
        ):
            return

        album = self.names.original_album.lower()
        if not (
            # Some tracks have no artist at all - reset makes sense
            self.tracks_without_artist
            # Parsed artist looks like album name prefix
            or any(album.startswith(a) for a in unique_artists)
        ):
            return

        # Merge the wrongly-parsed artist back into the title and restore the album
        # artist
        for t in tracks_with_parsed_artist:
            # Prepend the fragment back to the title
            t.title = f"{t.artist} - {t.title}"
            t.artist = albumartist  # Restore the correct album artist

    def for_media(
        self, media: str, comments: str, include_digi: bool
    ) -> list[dict[str, Any]]:
        if not include_digi and media != "Digital Media":
            tracks_ = [t for t in self if not t.digi_only]
        else:
            tracks_ = list(self)

        medium_total = {"medium_total": len(tracks_)}
        tracks = [t.info | medium_total for t in tracks_]
        if len(tracks) == 1 or media != "Vinyl":
            return tracks

        # using an ordered set here in case of duplicates
        track_alts = ordset(Track.TRACK_ALT_PAT.findall(comments))
        if len(track_alts) != len(tracks):
            return tracks

        mediums = [self.DISC_BY_LETTER[ta[0]] for ta in track_alts]
        total_by_medium = Counter(mediums)
        index_by_medium = dict.fromkeys(total_by_medium, 1)
        for track, (track_alt, medium) in zip(tracks, zip(track_alts, mediums)):
            track.update(
                track_alt=track_alt,
                medium=medium,
                medium_total=total_by_medium[medium],
                medium_index=index_by_medium[medium],
            )
            index_by_medium[medium] += 1

        return tracks



"""Module for parsing bandcamp metadata."""

from __future__ import annotations

import itertools as it
import json
import operator as op
import re
from datetime import datetime, timezone
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, ClassVar
from unicodedata import normalize

from beets import config as beets_config
from beets.autotag.hooks import AlbumInfo, TrackInfo
from pycountry import countries, subdivisions

from .album_name import AlbumName
from .catalognum import Catalognum
from .helpers import Helpers, MediaInfo, cached_patternprop
from .names import Names
from .tracks import Tracks

if TYPE_CHECKING:
    from collections.abc import Iterable
    from datetime import date


JSONDict = dict[str, Any]

COUNTRY_OVERRIDES = {
    "Russia": "RU",  # pycountry: Russian Federation
    "The Netherlands": "NL",  # pycountry: Netherlands
    "UK": "GB",  # pycountry: Great Britain
    "D.C.": "US",
    "South Korea": "KR",  # pycountry: Korea, Republic of
    "Turkey": "TR",  # pycountry: only handles Türkiye
}
DATA_SOURCE = "bandcamp"
WORLDWIDE = "XW"
DIGI_MEDIA = "Digital Media"
VA = "Various Artists"
VA_ARTIST_COUNT = 4  # this number of artists is replaced with VA name


class Metaguru(Helpers):
    META_PAT = cached_patternprop(r'.*"@id".*')
    LABEL_IN_DESC = cached_patternprop(r"(?<=Label:) *\b[^/,\n]+")
    ARTIST_IN_DESC = cached_patternprop(r"Artists?: *(\b[^\n]+)")
    REMIX_IN_ARTIST = cached_patternprop(r"(?:[(,+]|w/)+.*?re?mi?x", re.I)
    NOT_ALPHANUMERIC = cached_patternprop(r"\W")
    HTML_REMOVE_CHARS: ClassVar[list[str]] = [
        "\u200b",
        "\u200d",
        "\u200e",
        "\u200f",
        "\u00a0",
    ]

    _singleton = False
    va_name = VA
    media = MediaInfo("", "", "", "")

    meta: JSONDict
    config: JSONDict
    media_formats: list[MediaInfo]
    _tracks: Tracks
    _album_name: AlbumName

    def __init__(self, meta: JSONDict, config: JSONDict | None = None) -> None:
        self.meta = meta
        self.media_formats = self.get_media_formats(
            (meta.get("inAlbum") or meta).get("albumRelease") or []
        )
        if self.media_formats:
            self.media = self.media_formats[0]
        self.config = config or {}
        self.va_name = beets_config["va_name"].as_str() or self.va_name
        names = Names(meta, self.original_albumartist)
        names.resolve()
        self._names = names
        self._tracks = Tracks.from_names(names)
        self._catalognum = Catalognum(
            f"{self.description}\n{self.credits}".replace("\r", ""),
            names.original_album,
            names.label,
            [*self._tracks.artists_and_titles, self.original_albumartist],
        )
        self._album_name = AlbumName(
            names.original_album, self.all_media_comments, names.album_in_titles
        )

    @classmethod
    def from_html(cls, html: str, config: JSONDict | None = None) -> Metaguru:
        for char in cls.HTML_REMOVE_CHARS:
            html = html.replace(char, "")
        try:
            meta = cls.META_PAT.search(html).group()  # type: ignore[union-attr]
        except AttributeError as exc:
            raise AttributeError("Could not find release metadata JSON") from exc

        return cls(json.loads(meta), config)

    @cached_property
    def excluded_fields(self) -> set[str]:
        return set(self.config.get("exclude_extra_fields") or [])

    @cached_property
    def description(self) -> str:
        return self.meta.get("description") or ""

    @cached_property
    def credits(self) -> str:
        return self.meta.get("creditText") or ""

    @property
    def comments(self) -> str | None:
        """Return concatenated release, media descriptions and credits."""
        parts = [self.description]

        normalize = partial(self.NOT_ALPHANUMERIC.sub, "")
        if normalize(self.media.description.lower()) != normalize(
            self.description.lower()
        ):
            parts.append(self.media.description)
        parts.append(self.credits)

        sep = self.config["comments_separator"]
        return sep.join(filter(None, parts)).replace("\r", "") or None

    @cached_property
    def disctitles(self) -> str:
        return " ".join([m.disctitle for m in self.media_formats if m.disctitle])

    @cached_property
    def only_media_comments(self) -> str:
        return "\n".join(
            [
                self.disctitles,
                *[m.description for m in self.media_formats],
            ]
        )

    @cached_property
    def all_media_comments(self) -> str:
        return "\n".join([self.only_media_comments, self.comments or ""])

    @cached_property
    def label(self) -> str:
        if m := self.LABEL_IN_DESC.search(self.all_media_comments):
            return m[0].strip(" '\"")

        return self._names.label

    @cached_property
    def album_id(self) -> str:
        return self.meta.get("@id") or ""

    @cached_property
    def artist_id(self) -> str:
        try:
            return self.meta["byArtist"]["@id"]  # type: ignore [no-any-return]
        except KeyError:
            return self.meta["publisher"]["@id"]  # type: ignore [no-any-return]

    @cached_property
    def original_albumartist(self) -> str:
        if m := self.ARTIST_IN_DESC.search(self.all_media_comments):
            aartist = m[1].strip()
        else:
            aartist = self.meta["byArtist"]["name"]

        aartist = ", ".join(map(str.strip, aartist.split(" // ")))
        return self.REMIX_IN_ARTIST.split(aartist)[0].strip()

    @cached_property
    def original_album(self) -> str:
        return self._names.original_album

    @cached_property
    def preliminary_albumartist(self) -> str:
        """Determine and return preliminary album artist to set as default for tracks.

        This property calculates the lead artist based on the original album artist,
        label, and track track collaborators. Since this property gets required before
        track artists are available, it uses the pinciple of elimination to find the
        albumartist candidate.
        """
        aartist = self.original_albumartist
        if self.label != aartist:
            aartist = Helpers.clean_name(aartist)
        elif a := self._album_name.find_artist(self.catalognum):
            aartist = a

        if (
            len(aartists := Helpers.split_artists(aartist)) > 1
            and (main_artists := self._tracks.discard_collaborators(aartists))
            and main_artists != aartists
        ):
            return ", ".join(main_artists)

        return aartist

    @cached_property
    def cover_art_url(self) -> str:
        image = self.meta.get("image") or ""
        if isinstance(image, list) and isinstance(image[0], str):
            return image[0]
        return image

    @cached_property
    def release_date(self) -> date | None:
        """Parse the datestring that takes the format like below and return date object.

        {"datePublished": "17 Jul 2020 00:00:00 GMT"}.

        If the field is not found, return None.
        """
        if dt := self.meta.get("datePublished") or self.meta.get("dateModified"):
            return (
                datetime.strptime(dt[:11], "%d %b %Y")
                .replace(tzinfo=timezone.utc)
                .date()
            )

        return None

    @cached_property
    def albumstatus(self) -> str:
        reldate = self.release_date
        today = datetime.now(tz=timezone.utc).date()
        return "Official" if reldate and reldate <= today else "Promotional"

    @property
    def disctitle(self) -> str | None:
        """Return medium's disc title if found."""
        return self.media.disctitle or None

    @staticmethod
    def get_mediums(track_infos: list[TrackInfo]) -> int:
        """Get the count of discs / mediums for this release format."""
        return max((t.medium or 0 for t in track_infos), default=0)

    @property
    def catalognum(self) -> str:
        """Return the first found catalogue number."""
        return (
            self._names.catalognum
            or self._catalognum.get(f"{self.media.disctitle}\n{self.media.description}")
            or ""
        )

    @cached_property
    def country(self) -> str:
        try:
            loc = self.meta["publisher"]["foundingLocation"]["name"].rpartition(", ")[
                -1
            ]
            name = normalize("NFKD", loc).encode("ascii", "ignore").decode()
            return (
                COUNTRY_OVERRIDES.get(name)
                or getattr(countries.get(name=name, default=object), "alpha_2", None)
                or subdivisions.lookup(name).country_code
            )
        except (ValueError, LookupError):
            return WORLDWIDE

    @cached_property
    def tracks(self) -> Tracks:
        """Return parsed tracks."""
        self._tracks.fix_track_artists(self.preliminary_albumartist)
        return self._tracks

    @cached_property
    def unique_artists(self) -> list[str]:
        """Return all unique artists in the release ignoring differences in case."""
        artists = self.split_artists(self._tracks.artists)
        return artists[:1] if len(set(map(str.lower, artists))) == 1 else artists

    @cached_property
    def track_count(self) -> int:
        return len(self._tracks)

    @cached_property
    def albumartist(self) -> str:
        """Take into account the release contents and return the actual albumartist.

        If we have one track, return the artist of that track.
        If we have more than VA_ARTIST_COUNT artists, return the VA name.

        Compare the original albumartist with artists found in the tracks:
        * if we have a single albumartist, and it's one of the track lead artists,
          return it ignoring the rest of track artists
        * if we have more than one albumartist, and they are all track lead artists,
          return them in the original format
        * otherwise, join track lead artists with a comma and return them.

        Note: ft artists are not included.
        """
        if self.track_count == 1:
            return self.remove_ft(self.tracks.first.artist)

        if self.va:
            return self.va_name

        aartist = self.preliminary_albumartist
        if aartist and (set(self.split_artists(aartist)) <= set(self.unique_artists)):
            return self.remove_ft(aartist) if "remix" in aartist else aartist
        if len(self.tracks.original_artists) == 1:
            return self.tracks.original_artists[0]

        def normalize(artists: Iterable[str]) -> tuple[str, ...]:
            return tuple(sorted(set(map(str.lower, artists))))

        aartists = normalize(self.split_artists(aartist, force=True))
        all_artists_sets = {
            normalize(self._tracks.lead_artists),
            normalize(self.unique_artists),
        }
        all_artists = {a for artists in all_artists_sets for a in artists}
        if (
            not self.tracks.lead_artists
            # if the release specifies a single albumartist, respect it as long as
            # that's one of the track artists
            or (len(aartists) == 1 and next(iter(aartists)) in all_artists)
            or (aartists in all_artists_sets)
        ):
            return aartist

        return ", ".join(sorted(self.tracks.lead_artists))

    @property
    def album_name(self) -> str:
        artists = []
        if self.original_albumartist != self.label:
            artists.append(self.original_albumartist)

        return self._album_name.get(
            self.catalognum,
            [*artists, *self.tracks.original_artists],
            self.tracks.artists,
            self.label,
        )

    def _search_albumtype(self, word: str) -> bool:
        """Return whether the given word (ep or lp) matches the release albumtype.

        True when one of the following conditions is met:
        * if {word}[0-9] is found in the catalognum
        * if [0-9]?{word} is found in the album name or any disctitle
        * if {word} is preceded by 'this ...' or 'the' in a sentence in the release or
          or any media description, like 'this EP'
        * if {word} and the album_name are found in the same sentence in the release or
          any media description.
        """
        text = " ".join(self.all_media_comments.splitlines())
        sentences = [s.strip() for s in text.lower().split(". ")]

        word_pat = re.compile(rf"(?<!-)\b{word}(\b|\.)", re.I)
        in_catnum = re.compile(rf"{word}\d", re.I)
        release_ref = re.compile(
            rf"\b((this|with|present|deliver|new)[\w\s,'-]*?|the|track|full|first) {word}\b",  # noqa: E501
            re.I,
        )
        album_name = self.album_name.lower()

        media_word_pat = re.compile(rf"(vinyl |x|[0-5]){word}\b", re.I)
        return bool(
            media_word_pat.search(self.only_media_comments)
            or in_catnum.search(self.catalognum)
            or word_pat.search(f"{album_name} {self.disctitles}")
            or in_catnum.search(text)
            or any(
                release_ref.search(s) or (word_pat.search(s) and album_name in s)
                for s in sentences
            )
        )

    @cached_property
    def is_singleton(self) -> bool:
        return self._singleton or self.track_count == 1

    @cached_property
    def is_single_album(self) -> bool:
        return self.track_count > 1 and (
            len({t.title_without_remix for t in self.tracks}) == 1
        )

    @cached_property
    def is_lp(self) -> bool:
        """Return whether the release is an LP."""
        return self._search_albumtype("lp")

    @cached_property
    def is_ep(self) -> bool:
        """Return whether the release is an EP."""
        return self._search_albumtype(r"e\.?p") or (
            " / " in self.album_name
            and len(self.tracks.artists) == AlbumName.SPLIT_RELEASE_ARTIST_COUNT
        )

    @cached_property
    def is_comp(self) -> bool:
        """Return whether the release is a compilation."""
        return (
            self._album_name.mentions_compilation
            or self._search_albumtype("compilation")
            or (
                len(self.tracks.lead_artists) >= VA_ARTIST_COUNT
                and self.track_count > VA_ARTIST_COUNT
            )
        )

    @cached_property
    def albumtype(self) -> str:
        if self.is_singleton:
            return "single"
        if self.is_lp:
            return "album"
        if self.is_ep:
            return "ep"
        if self.is_single_album:
            return "single"
        return "compilation" if self.is_comp else "album"

    @cached_property
    def albumtypes(self) -> list[str]:
        albumtypes = {self.albumtype}
        if self.is_comp:
            if self.albumtype == "ep":
                albumtypes.add("compilation")
            else:
                albumtypes.add("album")
        if not self.is_singleton and self.is_lp:
            albumtypes.add("lp")
        if self.is_single_album:
            albumtypes.add("single")

        if self.albumtype == "single" and self.track_count > 1:
            albumtypes.add("album")

        for word in ["remix", "rmx", "edits", "live", "soundtrack"]:
            if word in self.original_album.lower():
                albumtypes.add(word.replace("rmx", "remix").replace("edits", "remix"))

        if sum(bool(t.remix and t.remix.valid) for t in self._tracks) >= max(
            self.track_count - 1, 1
        ):
            albumtypes.add("remix")

        return sorted(albumtypes)

    @cached_property
    def va(self) -> bool:
        return len(self.tracks.lead_artists) >= VA_ARTIST_COUNT

    @cached_property
    def style(self) -> str | None:
        """Extract bandcamp genre tag from the metadata."""
        # expecting the following form: https://bandcamp.com/tag/folk
        tag_url = self.meta.get("publisher", {}).get("genre") or ""
        style = None
        if tag_url:
            style = tag_url.split("/")[-1]
            if self.config["genre"]["capitalize"]:
                style = style.capitalize()
        return style

    @cached_property
    def genres(self) -> list[str] | None:
        kws: Iterable[str] = map(str.lower, self.meta.get("keywords", []))
        if self.style:
            exclude_style = partial(op.ne, self.style.lower())
            kws = filter(exclude_style, kws)

        genre_cfg = self.config["genre"]
        genres = self.get_genre(kws, genre_cfg, self.label)
        if genre_cfg["capitalize"]:
            genres = map(str.capitalize, genres)
        if genre_cfg["maximum"]:
            genres = it.islice(genres, genre_cfg["maximum"])

        return sorted(genres) or None

    @cached_property
    def artists(self) -> list[str]:
        artists = self.split_artists(self.albumartist, force=True)
        if m := self.FT_PAT.search(self.albumartist):
            artists.append(m["ft_artist"])

        return artists

    @property
    def _common(self) -> JSONDict:
        return {
            "album": self.album_name,
            "artist_id": self.artist_id,
            "artists_credit": [],  # TODO: implement
            "artists_ids": [self.artist_id],
            "artists_sort": [],  # TODO: implement
            "data_source": DATA_SOURCE,
            "data_url": self.album_id,
            "media": self.media.name,
        }

    def get_fields(self, fields: list[str], src: object = None) -> JSONDict:
        """Return a mapping between unexcluded fields and their values."""
        return dict(zip(fields, iter(op.attrgetter(*fields)(src or self))))

    @property
    def _common_album(self) -> JSONDict:
        common_data: JSONDict = dict.fromkeys(["barcode", "release_group_title"])
        fields = [
            "albumtype",
            "albumtypes",
            "catalognum",
            "comments",
            "country",
            "genres",
            "label",
            "style",
            "artists",
            "cover_art_url",
        ]
        common_data.update(self.get_fields(fields))
        if reldate := self.release_date:
            common_data.update(self.get_fields(["year", "month", "day"], reldate))
            common_data["original_year"] = reldate.year
            common_data["original_month"] = reldate.month
            common_data["original_day"] = reldate.day

        return common_data

    def _exclude_extra_fields(self, data: JSONDict) -> JSONDict:
        return {k: v for k, v in data.items() if k not in self.excluded_fields}

    def _trackinfo(self, data: dict[str, Any], **kwargs: Any) -> TrackInfo:
        data.update(**self._common, **kwargs)
        # if track-level catalognum is not found or if it is the same as album's, then
        # remove it. Otherwise, keep it attached to the track
        if not data["catalognum"] or data["catalognum"] == self.catalognum:
            data.pop("catalognum", None)
        if not data["lyrics"]:
            data.pop("lyrics", None)

        return TrackInfo(**self._exclude_extra_fields(data))

    @cached_property
    def singleton(self) -> TrackInfo:
        self._singleton = True
        self.media = self.media_formats[0]
        track_data = self.tracks.first.info
        catalognum = track_data["catalognum"] or self.catalognum
        data = {
            **track_data,
            **self._common_album,
            "medium": None,
            "track_id": self.album_id,
            "title": track_data["title"] or catalognum,
        }
        track = self._trackinfo(data)
        track.album = None
        track.catalognum = catalognum
        return self.check_list_fields(track)

    def get_media_album(self, media: MediaInfo) -> AlbumInfo:
        """Return album for the appropriate release format."""
        self.media = media

        tracks = self.tracks.for_media(
            self.media.name,
            self.comments or "",
            bool(self.config.get("include_digital_only_tracks")),
        )
        track_infos = [self._trackinfo(t, disctitle=self.disctitle) for t in tracks]
        data = {
            **self._common,
            **self._common_album,
            "artist": self.albumartist,
            "mediums": self.get_mediums(track_infos),
            "albumstatus": self.albumstatus,
            "va": self.va,
            "tracks": track_infos,
            "album_id": self.media.album_id,
        }
        return self.check_list_fields(AlbumInfo(**self._exclude_extra_fields(data)))

    @cached_property
    def albums(self) -> list[AlbumInfo]:
        """Return album for the appropriate release format."""
        return list(map(self.get_media_album, self.media_formats))



import atexit
import re
from functools import cache, partial
from html import unescape

import beets
import httpx

USER_AGENT = f"beets/{beets.__version__} +https://beets.io/"

_client = httpx.Client(headers={"User-Agent": USER_AGENT})

_rm_single_quote_dot = partial(re.compile(r"'|(?<=\d)\.(?=\d)").sub, "")
_non_ascii_to_dash = partial(re.compile(r"\W", flags=re.ASCII).sub, "-")
_squeeze_dashes = partial(re.compile(r"--+").sub, "-")


@atexit.register
def close_client() -> None:
    """Close the http client at exit."""
    _client.close()


@cache
def http_get_text(url: str) -> str:
    """Return text contents of the url."""

    response = _client.get(url, follow_redirects=True)
    response.raise_for_status()

    return unescape(response.text)


def urlify(pretty_string: str) -> str:
    """Transform a string into bandcamp url."""
    name = _rm_single_quote_dot(pretty_string.lower())
    return _squeeze_dashes(_non_ascii_to_dash(name)).strip("-")



"""Module with a single track parsing functionality."""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Any

from .catalognum import Catalognum
from .helpers import Helpers, cached_patternprop

if TYPE_CHECKING:
    from .helpers import JSONDict

digiwords = r"""
    # must contain at least one of
    ([ -]?  # delimiter
        (bandcamp|digi(tal)?|exclusive|bonus²?|bns|unreleased)
    )+
    # and may be followed by
    (\W(tr?ack|only|tune|bootleg))*
    """


@dataclass
class Remix:
    PATTERN = cached_patternprop(
        r"""
    (?P<start>^)?
    \ *\(?
    (?P<text>
      (?:
          (?P<b>\[)
        | (?P<p>\((?!.*\())
        | (?<!-)-\ (?!.*([([]|\ -\ ))
      )
      (?P<remixer>['"]?\b\w.*?|)\ *
      (?P<type>(re)?mix|rmx|edit|bootleg|(?<=\w\ )version|remastered)\b
      [^])]*
      (?(b)\])
      (?(p)\))
    )
    (?P<end>$)?
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    full: str
    remixer: str
    text: str
    type: str
    start: bool
    end: bool

    @classmethod
    def from_name(cls, name: str) -> Remix | None:
        if m := cls.PATTERN.search(name):
            remix: dict[str, Any] = m.groupdict()
            remix["start"] = remix["start"] is not None
            remix["end"] = remix["end"] is not None
            remix["type"] = remix["type"].lower()
            remix.pop("b")
            remix.pop("p")
            return cls(**remix, full=m[0])
        return None

    @cached_property
    def valid(self) -> bool:
        return self.remixer.lower() != "original" and self.type != "remastered"

    @cached_property
    def artist(self) -> str | None:
        if self.valid and self.remixer.lower() != "extended" and self.type != "version":
            return self.remixer

        return None


@dataclass
class Track:
    DIGI_ONLY_PAT = cached_patternprop(
        rf"""
    (\s|[^][()\w])*  # space or anything that is not a parens or an alphabetical char
    (
             (^{digiwords}[.:\d\s]+\s)  # begins with 'Bonus.', 'Bonus 1.' or 'Bonus :'
     | [\[(]\s*{digiwords}\s*[\])]\W*   # delimited by brackets, '[Bonus]', '(Bonus) -'
     |      [*]{digiwords}[*]?          # delimited by asterisks, '*Bonus', '*Bonus*'
     |         {digiwords}[ ]-          # followed by ' -', 'Bonus -'
     |     ([ ]{digiwords}$)            # might not be delimited at the end, '... Bonus'
    )
    \s*  # all succeeding space
        """,
        re.I | re.VERBOSE,
    )
    DELIM_NOT_INSIDE_PARENS = cached_patternprop(
        r"(?<!-)(?<!^live)(?<!sample\ pack) - (?!-|[^([]+\w[])])", re.I
    )
    TRACK_ALT_PAT = cached_patternprop(
        r"""
        (?:(?<=^)|(?<=-\ ))             # beginning of the line or after the separator
        (
            (?:[A-J]{1,3}[12]?\.?\d)    # A1, B2, E4, A1.1 etc.
          | (?:[AB]+(?!\ \()(?=\W{2}\b))# A, AA BB
        )
        (?:[/.:)_\s-]+)                 # consume the non-word chars for removal
        """,
        re.M | re.VERBOSE,
    )
    NUMBER_PAT = cached_patternprop(r"\d+")
    ARTIST_DELIM_PREFIX = cached_patternprop(r"^(and|x|\W+)+\b")

    json_item: JSONDict = field(default_factory=dict, repr=False)
    track_id: str = ""
    index: int | None = None
    medium_index: int | None = None
    json_artist: str = ""

    name: str = ""
    ft: str = ""
    catalognum: str | None = None
    ft_artist: str = ""
    remix: Remix | None = None

    digi_only: bool = False
    track_alt: str | None = None
    album_artist: str | None = None

    @classmethod
    def clean_digi_name(cls, name: str) -> tuple[str, bool]:
        """Clean the track title from digi-only artifacts.

        Return the clean name, and whether this track is digi-only.
        """
        clean_name = cls.DIGI_ONLY_PAT.sub("", name)
        return clean_name, clean_name != name

    @staticmethod
    def split_ft(value: str) -> tuple[str, str, str]:
        """Return ft artist, full ft string, and the value without the ft string."""
        if m := Helpers.FT_PAT.search(value):
            grp = m.groupdict()
            return grp["ft_artist"], grp["ft"], value.replace(m.group(), "")

        return "", "", value

    @classmethod
    def get_featuring_artist(cls, name: str, artist: str) -> dict[str, str]:
        """Find featuring artist in the track name.

        If the found artist is contained within the remixer, do not do anything.
        If the found artist is among the main artists, remove it from the name but
        do not consider it as a featuring artist.
        Otherwise, strip brackets and spaces and save it in the 'ft' field.
        """
        ft_artist, ft, name = cls.split_ft(name)

        if not ft_artist:
            ft_artist, ft, artist = cls.split_ft(artist)

        return {"name": name, "json_artist": artist, "ft": ft, "ft_artist": ft_artist}

    @classmethod
    def parse_name(cls, name: str, artist: str, index: int | None) -> JSONDict:
        result: JSONDict = {}
        artist, artist_digi_only = cls.clean_digi_name(artist)
        name, name_digi_only = cls.clean_digi_name(name)
        result["digi_only"] = name_digi_only or artist_digi_only

        if artist:
            artist = Helpers.clean_name(artist)
        name = Helpers.clean_name(name).strip()

        if m := cls.TRACK_ALT_PAT.search(name):
            result["track_alt"] = m.group(1).replace(".", "").upper()
            name = name.replace(m.group(), "")

        if m := Catalognum.delimited.search(name):
            result["catalognum"] = m.group(1)
            name = name.replace(m.group(), "").strip()

        # Remove leading index
        if index:
            name = re.sub(rf"^0?{index}\W\W+", "", name)
            result["medium_index"] = index

        if remix := Remix.from_name(name):
            result["remix"] = remix
            if remix.start:
                name = name.removeprefix(remix.full).strip()
            elif remix.end:
                name = name.removesuffix(remix.full).strip()

        return {**result, **cls.get_featuring_artist(name, artist)}

    @classmethod
    def make(cls, json: JSONDict) -> Track:
        artist = json.get("byArtist", {}).get("name", "")
        index = json.get("position")
        data = {
            "json_item": json,
            "track_id": json["@id"],
            "index": index,
            "album_artist": json.get("album_artist"),
            **cls.parse_name(json["name"], artist, index),
        }
        return cls(**data)

    @cached_property
    def duration(self) -> int | None:
        try:
            h, m, s = map(int, self.NUMBER_PAT.findall(self.json_item["duration"]))
        except KeyError:
            return None

        return h * 3600 + m * 60 + s

    @cached_property
    def lyrics(self) -> str:
        try:
            text: str = self.json_item["recordingOf"]["lyrics"]["text"]
        except KeyError:
            return ""

        return text.replace("\r", "")

    @cached_property
    def name_split(self) -> list[str]:
        name = self.name
        if (a := self.json_artist) and name.lower().startswith(
            artist_start := f"{a.lower()} - "
        ):
            return [name[len(artist_start) :]]

        split = self.DELIM_NOT_INSIDE_PARENS.split(name.strip())
        if self.json_artist and " - " not in name:
            return [self.json_artist.strip(), *split]

        return split

    @cached_property
    def title_without_remix(self) -> str:
        return self.name_split[-1]

    @cached_property
    def title(self) -> str:
        """Return the main title with the full remixer part appended to it."""
        if self.remix and self.remix.text not in self.title_without_remix:
            return f"{self.title_without_remix} {self.remix.text}"
        return self.title_without_remix

    @classmethod
    def clean_duplicate_artists(cls, artist: str, remix_text: str) -> str:
        """Remove the artist from the artist field if it's already in the remix text."""
        for subartist in Helpers.split_artists(artist, force=True):
            if subartist.lower() in remix_text:
                artist = re.sub(
                    rf"(and|x|\W+)*\b{re.escape(subartist)}", "", artist, flags=re.I
                )
                artist = cls.ARTIST_DELIM_PREFIX.sub("", artist)
        return artist

    @cached_property
    def artist(self) -> str:
        """Deduce the artist from the track name."""
        if self.album_artist:
            return self.album_artist

        if not self.title_without_remix:
            return ""

        if self.json_artist and len(self.name_split) == 1:
            return self.json_artist

        artist = " - ".join(self.name_split[:-1])
        initial_artist = artist
        artist = Remix.PATTERN.sub("", artist.strip(", "))
        if artist and self.remix and self.remix.artist:
            artist = self.clean_duplicate_artists(artist, self.remix.text.lower())

        # reset the artist back to the original for singletons, if it's gone
        if not artist and not self.index:
            artist = initial_artist

        return ", ".join(map(str.strip, artist.strip(", ").split(",")))

    @property
    def artists(self) -> list[str]:
        return Helpers.split_artists(self.artist)

    @property
    def lead_artist(self) -> str:
        if artists := Helpers.split_artists(self.artist, force=True):
            return artists[0]

        return self.artist

    @property
    def info(self) -> JSONDict:
        artists = self.artists
        if ft := self.ft_artist:
            artists.append(ft)
        artists = Helpers.split_artists(artists, force=True)

        return {
            "index": self.index,
            "medium_index": self.medium_index,
            "medium": 1,
            "track_id": self.track_id,
            "artist": (
                f"{self.artist} {self.ft}"
                if self.ft_artist not in self.artist + self.title
                else self.artist
            ),
            "artists": artists,
            "title": self.title,
            "length": self.duration,
            "track_alt": self.track_alt,
            "lyrics": self.lyrics,
            "catalognum": self.catalognum or None,
        }



"""Module with album parsing logic."""

from __future__ import annotations

import re
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any

from .helpers import Helpers, cached_patternprop
from .track import Remix, Track

if TYPE_CHECKING:
    from collections.abc import Iterable
    from re import Match

JSONDict = dict[str, Any]


@dataclass
class AlbumName:
    SPLIT_RELEASE_ARTIST_COUNT = 2

    _series = r"(?i:\b(part|volume|pt|vol)\b\.?)"
    SERIES = cached_patternprop(rf"{_series}[ ]?[A-Z\d.-]+\b")
    SERIES_FMT = cached_patternprop(rf"^(.+){_series} *0*")
    REMIX_IN_TITLE = cached_patternprop(
        r"""
            (?<=remixes\ )\([^()]+\)$
          | \((?:inc|\+)[^()]*mix(?:es)?\)
          | (?:incl\.|with\ remixes)[^()+]+
          | \W*(?:\+|w/)[\w\s/]*remix$
          | \(tracks\ from[^)]+\)
        """,
        re.IGNORECASE | re.VERBOSE,
    )
    CLEAN_EPLP = cached_patternprop(
        r"(?:[([]|Double ){0,2}(\b[EL]P\b)((,.*)?[])])?", re.I
    )
    ALBUM_SEP = cached_patternprop(r" [-:l|] |: |(?<=\] )")
    EPLP_ALBUM = cached_patternprop(r"\b.+\b[EL]P\b([ #\w]+$)?")
    EPLP_ALBUM_LINE = cached_patternprop(
        r"\b(?=[A-Z])(((?!Vinyl|VA|-)[^:\s]+ )+)[EL]P$", re.M
    )
    QUOTED_ALBUM = cached_patternprop(r"\B(['\"])([^'\"]+)\1\B( VA\d+| [EL]P)?", re.I)
    ALBUM_IN_DESC = cached_patternprop(r"(?:Title *: ?|Album(?: *:|/Single) )([^\n]+)")
    CLEAN_VA_EXCLUDE = cached_patternprop(r"\w various artists \w", re.I)
    CLEAN_VA = cached_patternprop(
        r"""(^v[./]?a|\W*Various(?:\ Artists?)?)\b(?!\ [A-z])[^A-z(]*""",
        re.IGNORECASE,
    )
    COMPILATION_IN_TITLE = cached_patternprop(r"compilation|best of|anniversary", re.I)
    YEAR_RANGE = cached_patternprop(r"20[12]\d - 20[12]\d")

    original: str
    description: str
    from_track_titles: str | None

    remove_artists = True

    @cached_property
    def from_description(self) -> str | None:
        """Try finding album name in the release description."""
        if m := self.ALBUM_IN_DESC.search(self.description):
            self.remove_artists = False
            return m.group(1).strip()

        return None

    @cached_property
    def mentions_compilation(self) -> bool:
        return bool(self.COMPILATION_IN_TITLE.search(self.original))

    @cached_property
    def from_title(self) -> str | None:
        """Try to guess album name from the original title.

        Return the first match from below, defaulting to None:
        1. If quotes are used in the title, they probably contain the album name.
        2. If 'EP' or 'LP' is in the original name, album name is what precedes it.
        """
        if m := self.QUOTED_ALBUM.search(self.original):
            return m.expand(r"\2\3")

        for part in self.ALBUM_SEP.split(self.original):
            if m := self.EPLP_ALBUM.search(part):
                return m.group()

        return None

    @cached_property
    def album_names(self) -> list[str]:
        priority_list = [
            self.from_track_titles,
            self.from_description,
            self.from_title,
            self.original,
        ]
        return list(filter(None, priority_list))

    @cached_property
    def name(self) -> str:
        return next(iter(self.album_names), "")

    @cached_property
    def series_part(self) -> str | None:
        """Return series if it is found in any of the album names."""
        for name in self.album_names:
            if m := self.SERIES.search(name):
                return m.group()

        return None

    @staticmethod
    def format_series(m: Match[str]) -> str:
        """Format the series part in an album.

        * Ensure 'Vol' or 'Pt' is suffixed by '.'
        * Ensure that it is followed by a space
        * Leading zeroes from the number are already removed by 'SERIES_FMT'.
        """
        album, series = m.groups()
        if album[0].isupper() and not series[0].isupper():
            series = series.capitalize()

        suffix = "." if len(series) in {2, 3} else ""
        return f"{album}{series}{suffix} "

    def standardize_series(self, album: str) -> str:
        """Standardize 'Vol', 'Part' etc. format."""
        if not (series := self.series_part):
            return album

        if series.lower() not in album.lower():
            # series found in the original name was not given in the description, but
            album += f", {series}"
        else:
            # move from the beginning to the end of the album
            album, moved = re.subn(rf"^({series})\W+(.+)", r"\2, \1", album)
            if not moved:
                # otherwise, ensure that it is delimited by a comma
                album = re.sub(rf"(?<=\w)( {series}(?!\)))", r",\1", album)

        return self.SERIES_FMT.sub(self.format_series, album)

    @classmethod
    def remove_va(cls, name: str) -> str:
        if not cls.CLEAN_VA_EXCLUDE.search(name):
            return cls.CLEAN_VA.sub(" ", name)

        return name

    @staticmethod
    def remove_pattern(name: str, pattern: str) -> str:
        allowed_chars = r"[*|,. \u2013\u2020]"
        return re.sub(
            rf"""
            (
                (?P<br>[([])                # match either an opening bracket/parens
              | (^|{allowed_chars})+        # and line start or any of the allowed chars
            )
            {pattern}
            (?(br)[])]                      # match closing brackets if they were opened
              | ({allowed_chars}|-|\d+$)*   # otherwise remove any of these chars
            )
            """,
            " ",
            name,
            flags=re.VERBOSE | re.IGNORECASE,
        ).strip("_: -")

    @classmethod
    def remove_label(cls, name: str, label: str) -> str:
        return cls.remove_pattern(
            name,
            rf"""
            (?<!\w\ )           # cannot be preceded by a simple word
            (?<!vs\.\ )         # cannot be preceded by vs.
            \b{re.escape(label)}\b
            (?!:\ Vol)          # cannot be followed by ': Vol'
            (?!.[&#A-z])        # cannot be followed by a word
            """,
        )

    @classmethod
    def remove_artist(cls, name: str, artist: str) -> str:
        artist = re.escape(artist).replace(r",\ ", r"(,\ |\ [x&]\ )")
        return cls.remove_pattern(
            name,
            rf"""
            (?<!\ [&,]\ )                       # keep B in 'A & B', 'A, B'
            ((compiled|selected)\ )?            # remove these prefixes if present
            (
                     (?<![\w.]\ ){artist}       # cannot be preceded by a word
              | ((by|split\ w)\ ){artist}       # unless preceded by 'by ' or 'split w '
            )
            (\ x\ [^-]+)?                       # remove other artist ' x C' if present
            (?![':,.\w])                        # cannot be followed by these characters
            (?!\ [a-wyz&])                      # cannot be followed by ' &' ' ep'
            """,
        )

    @classmethod
    def remove_catalognum(cls, name: str, catalognum: str) -> str:
        catalognum = re.escape(catalognum)
        with_space = re.sub(r"(?<=[A-Z])(?=\d)", " ", catalognum)
        return cls.remove_pattern(
            name, rf"/*(?i:{'|'.join({catalognum, with_space})})(?!\ deluxe)"
        )

    @classmethod
    def clean(
        cls,
        name: str,
        artists: list[str] | None = None,
        catalognum: str | None = None,
        label: str | None = None,
    ) -> str:
        """Return clean album name.

        Catalogue number and artists to be removed are provided as 'to_clean'.
        """
        name = Helpers.remove_ft(name)
        if catalognum:
            name = cls.remove_catalognum(name, catalognum)

        for artist in filter(None, artists or []):
            name = cls.remove_artist(name, artist)

        name = cls.remove_va(name)
        name = Helpers.clean_name(name)
        if label:
            name = cls.remove_label(name, label)
        name = cls.REMIX_IN_TITLE.sub(" ", name).strip("- ")

        # uppercase EP and LP, and remove surrounding parens / brackets
        name = cls.CLEAN_EPLP.sub(lambda x: x.group(1).upper(), name)
        return name.strip(" /")

    def find_artist(self, catalognum: str) -> str | None:
        album = self.original
        if self.YEAR_RANGE.match(album):
            return None

        album = self.clean(album, catalognum=catalognum)
        if remix := Remix.from_name(album):
            album = album.replace(remix.full, "").strip()

        if len(split := Track.DELIM_NOT_INSIDE_PARENS.split(album)) > 1:
            left, right = map(str.strip, split[:2])
            if ":" in left and self.SERIES.search(right):
                return None
            return left

        return None

    def check_eplp(self, album: str) -> str:
        """Return album name followed by 'EP' or 'LP' if that's given in the comments.

        When album is given, search for the album.
        Otherwise, search for (Capital-case Album Name) (EP or LP) and return the match.
        """
        if album:
            m = re.search(rf"{re.escape(album)} [EL]P\b", self.description)
        else:
            m = self.EPLP_ALBUM_LINE.search(self.description)

        return m.group() if m else album

    def get(
        self,
        catalognum: str,
        original_artists: list[str],
        artists: list[str],
        label: str,
    ) -> str:
        original_album = self.name
        if artists and original_album.lower() == artists[0].lower():
            # if album is named by the main artist, keep it as it is
            return original_album

        artists_to_clean: Iterable[str] = []
        if self.remove_artists:
            artists_to_clean = filter(None, original_artists + artists)
        artists_to_clean = sorted(set(artists_to_clean), key=len, reverse=True)

        album = self.clean(
            original_album, artists=artists_to_clean, catalognum=catalognum, label=label
        )
        if album.startswith("("):
            album = original_album

        album = self.check_eplp(self.standardize_series(album))

        if "split ep" in album.lower() or (
            not album and len(artists) == self.SPLIT_RELEASE_ARTIST_COUNT
        ):
            album = " / ".join(artists)

        return album or catalognum or original_album



"""Module with bandcamp search functionality."""
from __future__ import annotations

import re
from difflib import SequenceMatcher
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import quote_plus

from .http import http_get_text

if TYPE_CHECKING:
    from collections.abc import Callable

JSONDict = dict[str, Any]
SEARCH_URL = "https://bandcamp.com/search?page={}&q={}"


def _f(field: str) -> str:
    """Return pattern matching a string that does not start with '<' or space.

    Match until the end of the line.
    """
    return rf"(?P<{field}>[^\s<][^\n]+)"


RELEASE_PATTERNS = [
    re.compile(r"itemtype..\n\s+" + _f("type")),
    re.compile(r"search_item_type=[^>]+>\n\s+" + _f("name")),
    re.compile(r"\n\s+genre: " + _f("genre")),
    re.compile(r"\n\s+from " + _f("album")),
    re.compile(r"\n\s+by " + _f("artist")),
    re.compile(r"\n\s+released " + _f("date")),
    re.compile(r"\n\s+(?P<tracks>\d+) tracks"),
    re.compile(r">https://bandcamp\.(?P<label>[^.<]+)\.[^<]+<"),
    re.compile(r">https://(?P<label>[^.]+)\.bandcamp\.[^<]+<"),
    re.compile(r">https://(?P<label>(?!bandcamp)[^/]+)\.[^<]+<"),
    re.compile(r">(?P<url>https://[^<]+)<"),
]


def to_ascii(string: str) -> str:
    """Lowercase and translate non-ascii chars to '?'."""
    return string.lower().encode("ascii", "replace").decode()


def get_similarity(query: str, result: str) -> float:
    """Return the similarity between two strings normalized to [0, 1].

    We take into account how well the result matches the query, e.g.
        query: "foobar"
        result: "foo bar"
    Similarity is then:
        (2 * (len("foo") / len("foobar")) + len("foo") / len("foo bar")) / 3

    2/3 of the weight is how much of the query is found in the result,
    and 1/3 is a penalty for the non-matching part.
    """
    a, b = to_ascii(query), to_ascii(result)
    if not a or not b:
        return 0
    m = SequenceMatcher(a=a, b=b).find_longest_match(0, len(a), 0, len(b))
    return ((m.size / len(a)) * 2 + m.size / len(b)) / 3


def get_matches(text: str) -> JSONDict:
    """Reduce matches from all patterns into a single dictionary."""
    result: JSONDict = {}
    for pat in RELEASE_PATTERNS:
        if m := pat.search(text):
            result = {**m.groupdict(), **result}
    if "type" in result:
        result["type"] = result["type"].lower()
    if "date" in result:
        result["date"] = " ".join(reversed(result["date"].split()))
    return result


def parse_and_sort_results(html: str, **kwargs: str) -> list[JSONDict]:
    """Extract search results from `html` and sort them by similarity to kwargs.

    Bandcamp search may be unpredictable, therefore search results get sorted
    regarding their similarity to what's being queried.

    `kwargs` contains field and value pairs we compare the results with. Usually,
    this has 'label', 'artist' and 'name' ('title' or 'album') fields.
    """
    results: list[JSONDict] = []
    for block in html.split("searchresult data-search")[1:]:
        res = get_matches(block)
        similarities = [
            get_similarity(query, res.get(field, "")) for field, query in kwargs.items()
        ]
        res["similarity"] = round(sum(similarities) / len(similarities), 3)
        results.append(res)
    results = sorted(results, key=itemgetter("similarity"), reverse=True)
    return [{"index": i + 1, **r} for i, r in enumerate(results)]


def search_bandcamp(
    query: str = "",
    search_type: str = "",
    page: int = 1,
    get: Callable[[str], str] = http_get_text,
    **kwargs: Any,
) -> list[JSONDict]:
    """Return a list with item JSONs of type search_type matching the query."""
    query = query or " - ".join(
        filter(None, [kwargs.get("artist"), kwargs.get("name")])
    )
    kwargs.setdefault("name", query)
    url = SEARCH_URL.format(page, quote_plus(query))
    if search_type:
        url += f"&item_type={search_type}"
    return parse_and_sort_results(get(url), **kwargs)



"""List of MusicBrainz genres from https://beta.musicbrainz.org/genres."""

GENRES = {
    "2 tone",
    "2-step",
    "3-step",
    "aak",
    "abhang",
    "aboio",
    "abstract hip hop",
    "acholitronix",
    "acid breaks",
    "acid house",
    "acid jazz",
    "acid rock",
    "acid techno",
    "acid trance",
    "acidcore",
    "acousmatic",
    "acoustic blues",
    "acoustic chicago blues",
    "acoustic rock",
    "acoustic texas blues",
    "adhunik geet",
    "afoxê",
    "african blues",
    "afro house",
    "afro rock",
    "afro trap",
    "afro-cuban jazz",
    "afro-funk",
    "afro-jazz",
    "afro-zouk",
    "afrobeat",
    "afrobeats",
    "afropiano",
    "afroswing",
    "agbadza",
    "agbekor",
    "aggrotech",
    "ahwash",
    "aita",
    "akishibu-kei",
    "al jeel",
    "aleke",
    "algerian chaabi",
    "algorave",
    "alloukou",
    "alpenrock",
    "alté",
    "alternative country",
    "alternative dance",
    "alternative folk",
    "alternative hip hop",
    "alternative metal",
    "alternative pop",
    "alternative punk",
    "alternative r&b",
    "alternative rock",
    "amapiano",
    "ambasse bey",
    "ambient",
    "ambient americana",
    "ambient dub",
    "ambient house",
    "ambient noise wall",
    "ambient plugg",
    "ambient pop",
    "ambient techno",
    "ambient trance",
    "ambrosian chant",
    "american gamelan",
    "american primitive guitar",
    "americana",
    "amigacore",
    "anarcho-punk",
    "anatolian rock",
    "andalusian classical",
    "andean new age",
    "anglican chant",
    "animal sounds",
    "anti-folk",
    "aor",
    "apala",
    "appalachian folk",
    "aquacrunk",
    "arabesk",
    "arabesk rap",
    "arena rock",
    "arrocha",
    "arrocha funk",
    "arrocha sertanejo",
    "arrochadeira",
    "ars antiqua",
    "ars nova",
    "ars subtilior",
    "art pop",
    "art punk",
    "art rock",
    "art song",
    "artcore",
    "ashkenazi cantorial music",
    "asian rock",
    "asmr",
    "assiko",
    "atmospheric black metal",
    "atmospheric drum and bass",
    "atmospheric sludge metal",
    "audio documentary",
    "audio drama",
    "autonomic",
    "avant-folk",
    "avant-garde",
    "avant-garde jazz",
    "avant-garde metal",
    "avant-garde pop",
    "avant-prog",
    "avanzada",
    "avtorskaya pesnya",
    "axé",
    "bacardi",
    "bachata",
    "bachatón",
    "bagad",
    "bagatelle",
    "baguala",
    "baião",
    "baila",
    "baisha xiyue",
    "baithak gana",
    "bakersfield sound",
    "balani show",
    "balearic beat",
    "balearic trance",
    "balinese gamelan",
    "balitaw",
    "ballad",
    "ballad opera",
    "ballet",
    "ballet de cour",
    "ballroom house",
    "baltimore club",
    "bambuco",
    "banda sinaloense",
    "bandari",
    "bandinha",
    "banga",
    "barber beats",
    "barbershop",
    "bard rock",
    "bardcore",
    "baroque",
    "baroque pop",
    "baroque suite",
    "bashment soca",
    "bass house",
    "bassline",
    "batida",
    "batidão romântico",
    "batonebi songs",
    "battle rap",
    "battle record",
    "batucada",
    "batuque",
    "baul gaan",
    "beach music",
    "beat bolha",
    "beat bruxaria",
    "beat fino",
    "beat music",
    "beat poetry",
    "beatboxing",
    "beatdown hardcore",
    "bebop",
    "bedroom pop",
    "beijing opera",
    "bélé",
    "belgian techno",
    "bend-skin",
    "beneventan chant",
    "benga",
    "beni",
    "benna",
    "beompae",
    "bérite club",
    "berlin school",
    "bhajan",
    "bhangra",
    "bhavageethe",
    "bhojpuri pop",
    "big band",
    "big beat",
    "big room house",
    "big room trance",
    "biguine",
    "bikutsi",
    "binaural beats",
    "biraha",
    "birdsong",
    "birmingham sound",
    "bit music",
    "bitpop",
    "black 'n' roll",
    "black ambient",
    "black metal",
    "black midi",
    "black noise",
    "blackened crust",
    "blackened death metal",
    "blackgaze",
    "bleep techno",
    "blue-eyed soul",
    "bluegrass",
    "bluegrass gospel",
    "blues",
    "blues rock",
    "bocet",
    "boduberu",
    "boedra",
    "bogino duu",
    "bolero",
    "bolero español",
    "bolero son",
    "bolero-beat",
    "bomba",
    "bomba del chota",
    "bongo flava",
    "boogaloo",
    "boogie",
    "boogie rock",
    "boogie-woogie",
    "boom bap",
    "bossa nova",
    "bounce",
    "bounce beat",
    "bouncy techno",
    "bouyon",
    "brass band",
    "brazilian bass",
    "brazilian phonk",
    "break-in",
    "breakbeat",
    "breakbeat hardcore",
    "breakbeat kota",
    "breakcore",
    "breaks",
    "breakstep",
    "brega",
    "brega calypso",
    "brega funk",
    "briddim",
    "brill building",
    "brit funk",
    "britcore",
    "british blues",
    "british brass band",
    "british folk rock",
    "british rhythm & blues",
    "britpop",
    "bro-country",
    "broadband noise",
    "broken beat",
    "broken transmission",
    "brostep",
    "brutal death metal",
    "brutal prog",
    "bubblegum bass",
    "bubblegum dance",
    "bubblegum pop",
    "bubbling",
    "bubbling house",
    "buchiage trance",
    "budots",
    "bulería",
    "bullerengue",
    "burger-highlife",
    "burmese classical",
    "burmese mono",
    "burmese stereo",
    "burning spirits",
    "bytebeat",
    "byzantine chant",
    "c-pop",
    "c86",
    "ca trù",
    "cabaret",
    "cabo zouk",
    "cadence lypso",
    "cadence rampa",
    "cải lương",
    "cajun",
    "cakewalk",
    "čalgija",
    "calipso venezolano",
    "calypso",
    "campursari",
    "campus folk",
    "canción melódica",
    "candombe",
    "candombe beat",
    "cantata",
    "cante alentejano",
    "canterbury scene",
    "canto a lo poeta",
    "canto cardenche",
    "cantonese opera",
    "cantopop",
    "cantoria",
    "cantu a chiterra",
    "cantu a tenore",
    "canzona",
    "canzone d'autore",
    "canzone napoletana",
    "canzone neomelodica",
    "cape breton fiddling",
    "cape jazz",
    "caporal",
    "capriccio",
    "carimbó",
    "carnatic classical",
    "carnavalito",
    "carranga",
    "celtic",
    "celtic chant",
    "celtic electronica",
    "celtic metal",
    "celtic new age",
    "celtic punk",
    "celtic rock",
    "central asian throat singing",
    "chacarera",
    "chachachá",
    "chalga",
    "chamamé",
    "chamarrita açoriana",
    "chamarrita rioplatense",
    "chamber folk",
    "chamber pop",
    "champeta",
    "changa tuki",
    "change ringing",
    "changjak gugak",
    "changüí",
    "chanson à texte",
    "chanson française",
    "chanson québécoise",
    "chanson réaliste",
    "chaozhou xianshi",
    "chap hop",
    "character piece",
    "charanga",
    "chazzanut",
    "chèo",
    "chicago blues",
    "chicago bop",
    "chicago drill",
    "chicago house",
    "chicago soul",
    "chicano rap",
    "chicha",
    "children's music",
    "chilena",
    "chillout",
    "chillstep",
    "chillsynth",
    "chillwave",
    "chimayche",
    "chimurenga",
    "chinese classical",
    "chinese literati music",
    "chinese opera",
    "chinese revolutionary opera",
    "chipmunk soul",
    "chiptune",
    "chöd",
    "chopped and screwed",
    "choral symphony",
    "choro",
    "chotis madrileño",
    "christian hip hop",
    "christian metal",
    "christian rock",
    "christmas music",
    "church music",
    "chutney",
    "chutney soca",
    "cilokaq",
    "cinematic classical",
    "ciranda",
    "circus march",
    "city pop",
    "classic blues",
    "classic country",
    "classic jazz",
    "classic rock",
    "classical",
    "classical crossover",
    "classical period",
    "close harmony",
    "cloud rap",
    "club",
    "cocktail nation",
    "coco",
    "coladeira",
    "coldwave",
    "colindă",
    "colour bass",
    "comédie-ballet",
    "comedy",
    "comedy hip hop",
    "comedy rock",
    "comfy synth",
    "compas",
    "complextro",
    "concert band",
    "concertina band",
    "concerto",
    "concerto for orchestra",
    "concerto grosso",
    "conducted improvisation",
    "conga",
    "congolese rumba",
    "conscious hip hop",
    "contemporary christian",
    "contemporary classical",
    "contemporary country",
    "contemporary folk",
    "contemporary gospel",
    "contemporary jazz",
    "contemporary r&b",
    "contenance angloise",
    "contra",
    "cool jazz",
    "coon song",
    "copla",
    "corrido",
    "corrido tumbado",
    "cosmic country",
    "country",
    "country and irish",
    "country blues",
    "country boogie",
    "country folk",
    "country gospel",
    "country pop",
    "country rap",
    "country rock",
    "country soul",
    "country yodeling",
    "countrypolitan",
    "coupé-décalé",
    "cowboy poetry",
    "cowpunk",
    "crack rock steady",
    "crime jazz",
    "crossbreed",
    "crossover jazz",
    "crossover prog",
    "crossover thrash",
    "cruise",
    "crunk",
    "crunkcore",
    "crust punk",
    "csárdás",
    "cuarteto",
    "cubatón",
    "cuddlecore",
    "cueca",
    "cumbia",
    "cumbia amazónica",
    "cumbia argentina",
    "cumbia chilena",
    "cumbia colombiana",
    "cumbia mexicana",
    "cumbia norteña mexicana",
    "cumbia peruana",
    "cumbia pop",
    "cumbia rebajada",
    "cumbia santafesina",
    "cumbia sonidera",
    "cumbia turra",
    "cumbia villera",
    "cumbiatón",
    "cuplé",
    "currulao",
    "cyber metal",
    "cybergrind",
    "cyberpunk",
    "d-beat",
    "dabke",
    "dance",
    "dance-pop",
    "dance-punk",
    "dance-punk revival",
    "dance-rock",
    "dancefloor drum and bass",
    "dancehall",
    "dangak",
    "dangdut",
    "danmono",
    "dansband",
    "dansktop",
    "danzón",
    "dariacore",
    "dark ambient",
    "dark cabaret",
    "dark disco",
    "dark electro",
    "dark folk",
    "dark jazz",
    "dark plugg",
    "dark psytrance",
    "dark wave",
    "darkcore",
    "darkcore edm",
    "darkstep",
    "darksynth",
    "data sonification",
    "death 'n' roll",
    "death industrial",
    "death metal",
    "death-doom metal",
    "deathchant hardcore",
    "deathcore",
    "deathgrind",
    "deathrock",
    "deathstep",
    "dechovka",
    "deconstructed club",
    "deejay",
    "deep funk",
    "deep house",
    "deep soul",
    "deep tech",
    "deep techno",
    "delta blues",
    "dembow",
    "demostyle",
    "dennery segment",
    "denpa",
    "depressive black metal",
    "descarga",
    "desert blues",
    "desert rock",
    "desgarrada",
    "detroit techno",
    "detroit trap",
    "dhaanto",
    "dhol tasha",
    "dhrupad",
    "digicore",
    "digital cumbia",
    "digital fusion",
    "digital hardcore",
    "dikir barat",
    "dimotiko",
    "dirty south",
    "disco",
    "disco polo",
    "dissonant black metal",
    "dissonant death metal",
    "diva house",
    "divertissement",
    "dixieland",
    "djanba",
    "djent",
    "dobrado",
    "doina",
    "dondang sayang",
    "dongjing",
    "donk",
    "donosti sound",
    "doo-wop",
    "doom metal",
    "doomcore",
    "doomgaze",
    "doskpop",
    "downtempo",
    "downtempo deathcore",
    "dream pop",
    "dream trance",
    "dreampunk",
    "drift phonk",
    "drill",
    "drill and bass",
    "drone",
    "drone metal",
    "drum and bass",
    "drum and bugle corps",
    "drumfunk",
    "drumless hip hop",
    "drumline",
    "drumstep",
    "dub",
    "dub poetry",
    "dub techno",
    "dubstep",
    "dubstyle",
    "dubwise",
    "duma",
    "dunedin sound",
    "dungeon rap",
    "dungeon sound",
    "dungeon synth",
    "duranguense",
    "dutch house",
    "eai",
    "early hardstyle",
    "east coast hip hop",
    "easy listening",
    "easycore",
    "ebm",
    "eccojams",
    "edm",
    "electric blues",
    "electric texas blues",
    "electro",
    "electro hop",
    "electro house",
    "electro latino",
    "electro swing",
    "electro-disco",
    "electro-funk",
    "electro-industrial",
    "electroacoustic",
    "electroclash",
    "electronic",
    "electronic rock",
    "electronica",
    "electronicore",
    "electropop",
    "electropunk",
    "electrotango",
    "eleki",
    "eletrofunk",
    "embolada",
    "emo",
    "emo pop",
    "emo rap",
    "emocore",
    "emoviolence",
    "english pastoral school",
    "enka",
    "éntekhno",
    "epic collage",
    "epic doom metal",
    "estrada",
    "ethereal wave",
    "ethio-jazz",
    "étude",
    "euphoric hardstyle",
    "euro house",
    "euro-disco",
    "euro-trance",
    "eurobeat",
    "eurodance",
    "europop",
    "euskal kantagintza berria",
    "exotica",
    "experimental",
    "experimental big band",
    "experimental electronic",
    "experimental hip hop",
    "experimental rock",
    "expressionism",
    "extratone",
    "fado",
    "fado de coimbra",
    "fairy tale",
    "fakaseasea",
    "falak",
    "famo",
    "fandango",
    "fandango caiçara",
    "fantasia",
    "fantezi",
    "festejo",
    "festival progressive house",
    "festival trap",
    "fidget house",
    "field recording",
    "fife and drum",
    "fife and drum blues",
    "fijiri",
    "filin",
    "filipino rondalla",
    "filk",
    "filmi",
    "finnish tango",
    "flamenco",
    "flamenco jazz",
    "flamenco pop",
    "flashcore",
    "flex dance music",
    "florida breaks",
    "fm synthesis",
    "folk",
    "folk metal",
    "folk pop",
    "folk punk",
    "folk rock",
    "folkhop",
    "folktronica",
    "fon leb",
    "football chant",
    "footwork",
    "footwork jungle",
    "forest psytrance",
    "forró",
    "forró de favela",
    "forró eletrônico",
    "forró universitário",
    "frat rap",
    "frat rock",
    "freak folk",
    "freakbeat",
    "free car music",
    "free folk",
    "free improvisation",
    "free jazz",
    "free tekno",
    "freeform hardcore",
    "freestyle",
    "french electro",
    "french house",
    "frenchcore",
    "frevo",
    "frevo de bloco",
    "frevo de rua",
    "frevo elétrico",
    "frevo-canção",
    "fugue",
    "fuji",
    "full-on",
    "funaná",
    "funeral doom metal",
    "funeral march",
    "fungi",
    "funk",
    "funk automotivo",
    "funk brasileiro",
    "funk carioca",
    "funk de bh",
    "funk mandelão",
    "funk melody",
    "funk metal",
    "funk ostentação",
    "funk proibidão",
    "funk rock",
    "funknejo",
    "funkot",
    "funktronica",
    "funky breaks",
    "funky house",
    "fusion gugak",
    "future bass",
    "future bounce",
    "future core",
    "future funk",
    "future garage",
    "future house",
    "future rave",
    "future riddim",
    "futurepop",
    "futurism",
    "g-funk",
    "g-house",
    "gabber",
    "gaelic psalm singing",
    "gagaku",
    "gagok",
    "gaita zuliana",
    "gallican chant",
    "gambang kromong",
    "gamelan",
    "gamelan angklung",
    "gamelan beleganjur",
    "gamelan degung",
    "gamelan gender wayang",
    "gamelan gong gede",
    "gamelan gong kebyar",
    "gamelan jegog",
    "gamelan joged bumbung",
    "gamelan salendro",
    "gamelan sekaten",
    "gamelan selunding",
    "gamelan semar pegulingan",
    "gamelan siteran",
    "gamelan surakarta",
    "gangsta rap",
    "garage house",
    "garage psych",
    "garage punk",
    "garage rock",
    "garage rock revival",
    "garba",
    "geek rock",
    "género chico",
    "género grande",
    "genge",
    "gengetone",
    "għana",
    "ghazal",
    "ghetto funk",
    "ghetto house",
    "ghettotech",
    "ginan",
    "glam",
    "glam metal",
    "glam punk",
    "glam rock",
    "glitch",
    "glitch hop",
    "glitch hop edm",
    "glitch pop",
    "gnawa",
    "go-go",
    "goa trance",
    "gommance",
    "gondang",
    "goombay",
    "goregrind",
    "gorenoise",
    "gospel",
    "gospel house",
    "gospel reggae",
    "gothic",
    "gothic country",
    "gothic metal",
    "gothic rock",
    "gqom",
    "grand opera",
    "graphical sound",
    "grebo",
    "gregorian chant",
    "grime",
    "grindcore",
    "griot",
    "groove metal",
    "group sounds",
    "grunge",
    "grupera",
    "gstanzl",
    "guaguancó",
    "guajira",
    "guangdong yinyue",
    "guaracha",
    "guaracha edm",
    "guaracha santiagueña",
    "guarania",
    "gufeng",
    "guggenmusik",
    "guided meditation",
    "guitarrada",
    "gumbe",
    "guoyue",
    "gwo ka",
    "gypsy jazz",
    "gypsy punk",
    "habanera",
    "haitian vodou drumming",
    "halftime",
    "hambo",
    "hamburger schule",
    "hands up",
    "hanmai",
    "haozi",
    "hapa haole",
    "happy hardcore",
    "harana",
    "harawi",
    "hard beat",
    "hard bop",
    "hard drum",
    "hard house",
    "hard nrg",
    "hard rock",
    "hard techno",
    "hard trance",
    "hard trap",
    "hardbag",
    "hardbass",
    "hardcore breaks",
    "hardcore hip hop",
    "hardcore punk",
    "hardcore techno",
    "hardgroove techno",
    "hardstep",
    "hardstyle",
    "hardvapour",
    "hardwave",
    "harsh noise",
    "harsh noise wall",
    "hát tuồng",
    "hauntology",
    "heartland rock",
    "heaven trap",
    "heavy metal",
    "heavy psych",
    "heikyoku",
    "henan opera",
    "hexd",
    "hi-nrg",
    "hi-tech",
    "hi-tech full-on",
    "highlife",
    "hill country blues",
    "himene tarava",
    "hindustani classical",
    "hip hop",
    "hip hop soul",
    "hip house",
    "hipco",
    "hiplife",
    "holy minimalism",
    "honky tonk",
    "honkyoku",
    "hopepunk",
    "horror punk",
    "horror synth",
    "horrorcore",
    "house",
    "huapango",
    "huaylarsh",
    "huayno",
    "humppa",
    "hyangak",
    "hybrid trap",
    "hyper techno",
    "hyperpop",
    "hypertechno",
    "hyphy",
    "hypnagogic pop",
    "iavnana",
    "idm",
    "idol kayō",
    "illbient",
    "impressionism",
    "impromptu",
    "indeterminacy",
    "indian classical",
    "indian pop",
    "indie folk",
    "indie pop",
    "indie rock",
    "indie surf",
    "indietronica",
    "indo jazz",
    "indorock",
    "industrial",
    "industrial hardcore",
    "industrial hip hop",
    "industrial metal",
    "industrial musical",
    "industrial rock",
    "industrial techno",
    "instrumental",
    "instrumental hip hop",
    "instrumental jazz",
    "instrumental rock",
    "integral serialism",
    "interview",
    "iraqi maqam",
    "irish folk",
    "isa",
    "isicathamiya",
    "islamic modal music",
    "italo dance",
    "italo house",
    "italo-disco",
    "izlan",
    "izvorna bosanska muzika",
    "j-core",
    "j-pop",
    "j-rock",
    "jácara",
    "jackin house",
    "jaipongan",
    "jam band",
    "jamaican ska",
    "james bay fiddling",
    "jamgrass",
    "jangle pop",
    "japanese classical",
    "javanese gamelan",
    "jazz",
    "jazz blues",
    "jazz fusion",
    "jazz guachaca",
    "jazz house",
    "jazz mugham",
    "jazz poetry",
    "jazz pop",
    "jazz rap",
    "jazz rock",
    "jazz-funk",
    "jazzstep",
    "jeongak",
    "jerk",
    "jerk rap",
    "jersey club",
    "jersey club rap",
    "jersey drill",
    "jersey sound",
    "jesus music",
    "jiangnan sizhu",
    "jit",
    "jiuta",
    "joik",
    "jongo",
    "joropo",
    "jōruri",
    "jota",
    "jovem guarda",
    "jubilee",
    "jug band",
    "jùjú",
    "juke",
    "jump blues",
    "jump up",
    "jumpstyle",
    "jungle",
    "jungle dutch",
    "jungle terror",
    "junkanoo",
    "k-pop",
    "kabarett",
    "kacapi suling",
    "kadongo kamu",
    "kafi",
    "kagura",
    "kakawin",
    "kalattut",
    "kalindula",
    "kalon'ny fahiny",
    "kan ha diskan",
    "kaneka",
    "kankyō ongaku",
    "kantan chamorrita",
    "kanto",
    "kantruem",
    "kapuka",
    "kaseko",
    "kasékò",
    "kawaii future bass",
    "kawaii metal",
    "kayōkyoku",
    "kecak",
    "keroncong",
    "kete",
    "ketuk tilu",
    "khrueang sai",
    "khyal",
    "kidandali",
    "kidumbak",
    "kilapanga",
    "kirtan",
    "kizomba",
    "klapa",
    "klasik",
    "kleinkunst",
    "klezmer",
    "kliningan",
    "konnakol",
    "könsrock",
    "kontakion",
    "koplo",
    "korean ballad",
    "korean classical",
    "korean revolutionary opera",
    "kouta",
    "krakowiak",
    "krautrock",
    "kréyol djaz",
    "krushclub",
    "kuda lumping",
    "kuduro",
    "kujawiak",
    "kulintang",
    "kumi-daiko",
    "kumiuta",
    "kundiman",
    "kunqu",
    "kwaito",
    "kwassa kwassa",
    "kwela",
    "kyivan chant",
    "laiko",
    "lambada",
    "ländlermusik",
    "landó",
    "langgam jawa",
    "latin",
    "latin ballad",
    "latin disco",
    "latin funk",
    "latin house",
    "latin jazz",
    "latin pop",
    "latin rock",
    "latin soul",
    "lauda",
    "lavani",
    "lecture",
    "leftfield",
    "lento violento",
    "levenslied",
    "lied",
    "liedermacher",
    "liquid funk",
    "liscio",
    "livetronica",
    "liwa",
    "lo-fi",
    "lo-fi hip hop",
    "lo-fi house",
    "lolicore",
    "loner folk",
    "loud kei",
    "louisiana blues",
    "lounge",
    "lovers rock",
    "lowend",
    "lowercase",
    "luk krung",
    "luk thung",
    "lullaby",
    "lundu",
    "lute song",
    "mad",
    "madchester",
    "maddahi",
    "madrigal",
    "mafioso rap",
    "maftirim",
    "mahori",
    "mahraganat",
    "mainstream rock",
    "makina",
    "makossa",
    "malagueña venezolana",
    "malay gamelan",
    "malhun",
    "mallsoft",
    "malouf",
    "maloya",
    "maloya électronique",
    "maloya élektrik",
    "mambo",
    "mambo chileno",
    "mambo urbano",
    "mandopop",
    "manele",
    "mangambeu",
    "mangue beat",
    "manila sound",
    "mantra",
    "manyao",
    "marabi",
    "maracatu",
    "march",
    "marching band",
    "marchinha",
    "mariachi",
    "marinera",
    "marrabenta",
    "martial industrial",
    "mashcore",
    "maskanda",
    "mass",
    "mataali",
    "math pop",
    "math rock",
    "mathcore",
    "maxixe",
    "mazurka",
    "mbalax",
    "mbaqanga",
    "mbolé",
    "mbube",
    "mchiriku",
    "medieval",
    "medieval lyric poetry",
    "medieval metal",
    "medieval rock",
    "mega funk",
    "meiji shinkyoku",
    "melbourne bounce",
    "melodic bass",
    "melodic black metal",
    "melodic death metal",
    "melodic dubstep",
    "melodic hardcore",
    "melodic house",
    "melodic metalcore",
    "melodic techno",
    "melodic trance",
    "mélodie",
    "memphis rap",
    "mento",
    "menzuma",
    "merecumbé",
    "merengue",
    "merengue típico",
    "merenhouse",
    "merequetengue",
    "méringue",
    "merseybeat",
    "metal",
    "metalcore",
    "métis fiddling",
    "meyxana",
    "miami bass",
    "microfunk",
    "microhouse",
    "microsound",
    "microtonal classical",
    "midtempo bass",
    "midwest emo",
    "miejski folk",
    "milonga",
    "min'yō",
    "minatory",
    "mincecore",
    "minimal drum and bass",
    "minimal synth",
    "minimal techno",
    "minimal wave",
    "minimalism",
    "minneapolis sound",
    "minstrelsy",
    "mobb music",
    "mod",
    "mod revival",
    "moda de viola",
    "modal jazz",
    "modern blues",
    "modern classical",
    "modern creative",
    "modern hardtek",
    "modern laiko",
    "modinha",
    "moe song",
    "monodrama",
    "mood kayō",
    "moogsploitation",
    "moombahcore",
    "moombahton",
    "mor lam",
    "mor lam sing",
    "morenada",
    "morna",
    "moroccan chaabi",
    "motet",
    "motown",
    "moutya",
    "movimiento alterado",
    "mozarabic chant",
    "mpb",
    "mugham",
    "muiñeira",
    "mulatós",
    "muliza",
    "murga",
    "murga uruguaya",
    "musette",
    "music hall",
    "música cebolla",
    "música criolla",
    "música de intervenção",
    "música llanera",
    "música típica chilena",
    "musical",
    "mūsīqā lubnāniyya",
    "musique concrète",
    "musique concrète instrumentale",
    "muzika mizrahit",
    "muzika yehudit mekorit",
    "muzikat dika'on",
    "muziki wa dansi",
    "nagauta",
    "nanguan",
    "narcocorrido",
    "narodnozabavna glasba",
    "nasheed",
    "nashville sound",
    "native american new age",
    "nature sounds",
    "natya sangeet",
    "nederbeat",
    "nederpop",
    "neo kyma",
    "neo soul",
    "neo-acoustic",
    "neo-grime",
    "neo-medieval folk",
    "neo-progressive rock",
    "neo-psychedelia",
    "neo-rockabilly",
    "néo-trad",
    "neo-traditional country",
    "neoclassical dark wave",
    "neoclassical metal",
    "neoclassical new age",
    "neoclassicism",
    "neocrust",
    "neofolk",
    "neofolklore",
    "neon pop punk",
    "neoperreo",
    "nerdcore",
    "nerdcore techno",
    "neue deutsche härte",
    "neue deutsche welle",
    "neurofunk",
    "neurohop",
    "new age",
    "new beat",
    "new complexity",
    "new jack swing",
    "new jazz",
    "new mexico music",
    "new orleans blues",
    "new orleans r&b",
    "new rave",
    "new romantic",
    "new wave",
    "new york drill",
    "ngâm thơ",
    "ngoma",
    "nhạc đỏ",
    "nhạc tiền chiến",
    "nhạc vàng",
    "night full-on",
    "nightcore",
    "nigun",
    "nintendocore",
    "nitzhonot",
    "njuup",
    "no melody trap",
    "no wave",
    "nocturne",
    "noh",
    "noiadance",
    "noise",
    "noise pop",
    "noise rock",
    "noisecore",
    "noisegrind",
    "non-music",
    "nortec",
    "norteño",
    "northern soul",
    "nóta",
    "nouveau zydeco",
    "nova cançó",
    "novelty piano",
    "novo dub",
    "nu disco",
    "nu jazz",
    "nu metal",
    "nu skool breaks",
    "nu style gabber",
    "nueva canción",
    "nueva canción chilena",
    "nueva canción española",
    "nueva cumbia chilena",
    "nueva trova",
    "nuevo cancionero",
    "nuevo flamenco",
    "nuevo tango",
    "nustyle",
    "nwobhm",
    "nyū myūjikku",
    "oberek",
    "occult rock",
    "odissi classical",
    "ogene music",
    "oi",
    "old roman chant",
    "old school death metal",
    "old school hip hop",
    "old-time",
    "omutibo",
    "onda nueva",
    "ondō",
    "onkyo",
    "opera",
    "opera buffa",
    "opéra comique",
    "opera semiseria",
    "opera seria",
    "opera-ballet",
    "operatic pop",
    "operetta",
    "opm",
    "oratorio",
    "orchestral",
    "orchestral jazz",
    "orchestral song",
    "organic house",
    "ori deck",
    "oriental ballad",
    "orkes gambus",
    "orthodox pop",
    "outlaw country",
    "outsider house",
    "overture",
    "özgün müzik",
    "p-funk",
    "pachanga",
    "pacific reggae",
    "pagan black metal",
    "pagan folk",
    "paghjella",
    "pagodão",
    "pagode",
    "pagode romântico",
    "paisley underground",
    "palingsound",
    "palm-wine",
    "palo de mayo",
    "pansori",
    "parang",
    "parlour music",
    "partido alto",
    "pasillo",
    "pasodoble",
    "passion setting",
    "payada",
    "peak time techno",
    "pep band",
    "persian classical",
    "persian pop",
    "philly club",
    "philly club rap",
    "philly drill",
    "philly soul",
    "phleng phuea chiwit",
    "phonk",
    "phonk house",
    "piano blues",
    "piano rock",
    "picopop",
    "piedmont blues",
    "pigfuck",
    "pilón",
    "pimba",
    "pinpeat",
    "pìobaireachd",
    "pipe band music",
    "piphat",
    "pirekua",
    "piseiro",
    "piyyut",
    "pizzica",
    "plainchant",
    "plena",
    "plugg",
    "pluggnb",
    "plunderphonics",
    "poetry",
    "polca criolla",
    "political hip hop",
    "polka",
    "polka paraguaya",
    "polonaise",
    "pon-chak disco",
    "pop",
    "pop ghazal",
    "pop kreatif",
    "pop metal",
    "pop minang",
    "pop punk",
    "pop raï",
    "pop rap",
    "pop rock",
    "pop soul",
    "pop yeh-yeh",
    "porn groove",
    "pornogrind",
    "porro",
    "post-bop",
    "post-britpop",
    "post-classical",
    "post-dubstep",
    "post-grunge",
    "post-hardcore",
    "post-industrial",
    "post-metal",
    "post-minimalism",
    "post-punk",
    "post-punk revival",
    "post-rock",
    "powada",
    "power electronics",
    "power metal",
    "power noise",
    "power pop",
    "power soca",
    "powerstomp",
    "powerviolence",
    "praise & worship",
    "praise break",
    "prank calls",
    "prelude",
    "process music",
    "production music",
    "progressive",
    "progressive bluegrass",
    "progressive breaks",
    "progressive country",
    "progressive electronic",
    "progressive folk",
    "progressive house",
    "progressive metal",
    "progressive pop",
    "progressive psytrance",
    "progressive rock",
    "progressive soul",
    "progressive trance",
    "proto-punk",
    "psichedelia occulta italiana",
    "psybient",
    "psybreaks",
    "psychedelic",
    "psychedelic folk",
    "psychedelic pop",
    "psychedelic rock",
    "psychedelic soul",
    "psychobilly",
    "psychploitation",
    "psycore",
    "psystyle",
    "psytrance",
    "pub rock",
    "puirt à beul",
    "pumpcore",
    "pungmul",
    "punk",
    "punk blues",
    "punk rap",
    "punk rock",
    "punta",
    "punto",
    "purple sound",
    "puxa",
    "q-pop",
    "qaraami",
    "qasidah modern",
    "qawwali",
    "quan họ",
    "queercore",
    "quiet storm",
    "quyi",
    "r&b",
    "rabiz",
    "raga rock",
    "rage",
    "ragga",
    "ragga hip-hop",
    "ragga jungle",
    "raggacore",
    "raggatek",
    "ragtime",
    "raï",
    "rain sounds",
    "ranchera",
    "rap metal",
    "rap rock",
    "rapcore",
    "rapso",
    "raqs baladi",
    "rara",
    "rasin",
    "rasqueado cuiabano",
    "rasteirinha",
    "ratchet music",
    "rautalanka",
    "rave",
    "raw punk",
    "rawphoric",
    "rawstyle",
    "rebetiko",
    "red dirt",
    "red disco",
    "red song",
    "reductionism",
    "regalia",
    "reggae",
    "reggae rock",
    "reggae-pop",
    "reggaeton",
    "regional mexicano",
    "renaissance",
    "reparto",
    "repente",
    "requiem",
    "revue",
    "rhumba",
    "ricercar",
    "riddim dubstep",
    "rigsar",
    "ring shout",
    "riot grrrl",
    "ripsaw",
    "ritmada",
    "ritual ambient",
    "rizitika",
    "rkt",
    "rock",
    "rock and roll",
    "rock andaluz",
    "rock andino",
    "rock musical",
    "rock opera",
    "rock rural",
    "rock urbano",
    "rock urbano mexicano",
    "rockabilly",
    "rocksteady",
    "rōkyoku",
    "rom kbach",
    "romanian popcorn",
    "romantic classical",
    "romantic flow",
    "romantische oper",
    "rominimal",
    "roots reggae",
    "roots rock",
    "rumba",
    "rumba catalana",
    "rumba cubana",
    "rumba flamenca",
    "runo song",
    "russian chanson",
    "russian orthodox liturgical music",
    "russian romance",
    "ryūkōka",
    "sa'idi",
    "sacred harp",
    "sacred steel",
    "saeta",
    "salegy",
    "salsa",
    "salsa choke",
    "salsa dura",
    "salsa romántica",
    "saluang klasik",
    "samba",
    "samba de breque",
    "samba de gafieira",
    "samba de roda",
    "samba de terreiro",
    "samba soul",
    "samba-canção",
    "samba-choro",
    "samba-enredo",
    "samba-exaltação",
    "samba-jazz",
    "samba-joia",
    "samba-reggae",
    "samba-rock",
    "sambalanço",
    "sambass",
    "sample drill",
    "sampledelia",
    "samri",
    "sanjo",
    "santé engagé",
    "sarala gee",
    "sardana",
    "sarum chant",
    "sasscore",
    "sawt",
    "saya afroboliviana",
    "scam rap",
    "schlager",
    "schottische",
    "schranz",
    "scottish country dance music",
    "screamo",
    "scrumpy and western",
    "sea shanty",
    "sean-nós",
    "seapunk",
    "séga",
    "seggae",
    "seguidilla",
    "seishun punk",
    "semba",
    "semi-trot",
    "serenade",
    "serialism",
    "sermon",
    "sertanejo",
    "sertanejo raiz",
    "sertanejo romântico",
    "sertanejo universitário",
    "seto leelo",
    "sevdalinka",
    "sevillanas",
    "sexy drill",
    "shaabi",
    "shabad kirtan",
    "shan'ge",
    "shangaan electro",
    "shanto",
    "shashmaqam",
    "shatta",
    "shibuya-kei",
    "shidaiqu",
    "shima-uta",
    "shinkyoku",
    "shitgaze",
    "shoegaze",
    "shōmyō",
    "shoor",
    "sierreño",
    "sigidrigi",
    "sigilkore",
    "sinawi",
    "sinfonia concertante",
    "singeli",
    "singer-songwriter",
    "singspiel",
    "sitarsploitation",
    "sizhu music",
    "ska",
    "ska punk",
    "skacore",
    "skate punk",
    "sketch comedy",
    "skiffle",
    "skiladiko",
    "skinhead reggae",
    "skullstep",
    "skweee",
    "slack-key guitar",
    "slacker rock",
    "slam death metal",
    "slam poetry",
    "slap house",
    "sleaze rock",
    "slimepunk",
    "slow waltz",
    "slowcore",
    "sludge metal",
    "slushwave",
    "smooth jazz",
    "smooth soul",
    "snap",
    "soca",
    "soft rock",
    "sōkyoku",
    "son calentano",
    "son cubano",
    "son huasteco",
    "son istmeño",
    "son jarocho",
    "son montuno",
    "sonata",
    "songo",
    "sonorism",
    "sophisti-pop",
    "soukous",
    "soul",
    "soul blues",
    "soul jazz",
    "sound art",
    "sound collage",
    "sound effects",
    "sound poetry",
    "southeast asian classical",
    "southern gospel",
    "southern hip hop",
    "southern metal",
    "southern rock",
    "southern soul",
    "sovietwave",
    "space age pop",
    "space ambient",
    "space disco",
    "space rock",
    "space rock revival",
    "spacesynth",
    "spamwave",
    "spectralism",
    "speech",
    "speed garage",
    "speed house",
    "speed metal",
    "speedcore",
    "spiritual jazz",
    "spirituals",
    "splittercore",
    "spoken word",
    "spouge",
    "staïfi",
    "standup comedy",
    "steampunk",
    "steel band",
    "stenchcore",
    "sticheron",
    "stochastic music",
    "stomp and holler",
    "stoner metal",
    "stoner rock",
    "stornello",
    "street punk",
    "stride",
    "string quartet",
    "stutter house",
    "sufi rock",
    "sufiana kalam",
    "sundanese pop",
    "sungura",
    "sunshine pop",
    "suomisaundi",
    "surf",
    "surf punk",
    "surf rock",
    "sutartinės",
    "swamp blues",
    "swamp pop",
    "swamp rock",
    "swancore",
    "swing",
    "swing revival",
    "symphonic black metal",
    "symphonic metal",
    "symphonic mugham",
    "symphonic poem",
    "symphonic prog",
    "symphonic rock",
    "symphony",
    "synth funk",
    "synth-pop",
    "synthwave",
    "t-pop",
    "taarab",
    "tajaraste",
    "takamba",
    "talempong",
    "talempong goyang",
    "talking blues",
    "tallava",
    "tamborera",
    "tamborito",
    "tamborzão",
    "tammurriata",
    "tân cổ giao duyên",
    "tango",
    "tanjidor",
    "taoist ritual music",
    "tape music",
    "tappa",
    "taquirari",
    "tarana",
    "tarantella",
    "tarawangsa",
    "tarraxinha",
    "tassa",
    "tassu",
    "tchinkoumé",
    "tearout",
    "tearout brostep",
    "tech house",
    "tech trance",
    "technical death metal",
    "technical thrash metal",
    "techno",
    "techno bass",
    "techno kayō",
    "technobanda",
    "techstep",
    "tecnobrega",
    "tecnofunk",
    "tecnomerengue",
    "tecnorumba",
    "teen pop",
    "tejano",
    "tembang cianjuran",
    "terrorcore",
    "tex-mex",
    "texas blues",
    "texas country",
    "thai classical",
    "thall",
    "theme and variations",
    "third stream",
    "third wave ska",
    "thrash metal",
    "thrashcore",
    "thumri",
    "tibetan buddhist chant",
    "tiento",
    "timba",
    "timbila",
    "tin pan alley",
    "tivaner inngernerlu",
    "tizita",
    "toada de boi",
    "toccata",
    "tonada asturiana",
    "tonada potosina",
    "tonadilla",
    "tondero",
    "tontipop",
    "totalism",
    "township bubblegum",
    "township jive",
    "toypop",
    "toytown pop",
    "tradi-moderne congolais",
    "tradi-moderne ivoirien",
    "traditional black gospel",
    "traditional bluegrass",
    "traditional country",
    "traditional doom metal",
    "traditional pop",
    "tragédie en musique",
    "trallalero",
    "trampská hudba",
    "trance",
    "trance metal",
    "trancestep",
    "trap",
    "trap edm",
    "trap latino",
    "trap metal",
    "trap shaabi",
    "trap soul",
    "trapfunk",
    "tread",
    "tribal ambient",
    "tribal guarachero",
    "tribal house",
    "trikitixa",
    "trip hop",
    "troparion",
    "tropical house",
    "tropical rock",
    "tropicália",
    "tropicanibalismo",
    "tropipop",
    "trot",
    "trova",
    "trova yucateca",
    "truck driving country",
    "tsapiky",
    "tsonga disco",
    "tsugaru-jamisen",
    "tumba",
    "tumba francesa",
    "tumbélé",
    "turbo-folk",
    "turkish classical",
    "turntablism",
    "twee pop",
    "twerk",
    "twoubadou",
    "uaajeerneq",
    "udigrudi",
    "uk drill",
    "uk funky",
    "uk garage",
    "uk hardcore",
    "uk jackin",
    "uk street soul",
    "uk82",
    "unakesa",
    "underground hip hop",
    "unyago",
    "upopo",
    "uptempo hardcore",
    "urban contemporary gospel",
    "urban cowboy",
    "urtiin duu",
    "urumi melam",
    "us power metal",
    "utopian virtual",
    "uyghur muqam",
    "uzun hava",
    "v-pop",
    "vaigat",
    "vallenato",
    "vals criollo",
    "vals venezolano",
    "valsa brasileira",
    "vanera",
    "vapornoise",
    "vaportrap",
    "vaporwave",
    "vaudeville",
    "vaudeville blues",
    "vedic chant",
    "verbunkos",
    "verismo",
    "vietnamese bolero",
    "vietnamese classical",
    "viking metal",
    "viking rock",
    "villancico",
    "vinahouse",
    "visa",
    "visual kei",
    "vocal house",
    "vocal jazz",
    "vocal surf",
    "vocal trance",
    "vocalese",
    "volkstümliche musik",
    "vude",
    "wa euro",
    "waka",
    "waltz",
    "wangga",
    "war metal",
    "wassoulou",
    "waulking song",
    "wave",
    "weightless",
    "west coast breaks",
    "west coast hip hop",
    "west coast swing",
    "western",
    "western classical",
    "western swing",
    "whale song",
    "whistling",
    "white voice",
    "winter synth",
    "witch house",
    "wong shadow",
    "wonky",
    "wonky techno",
    "work song",
    "world fusion",
    "worldbeat",
    "wyrd folk",
    "xẩm",
    "xaxado",
    "xian psych",
    "xote",
    "xuc",
    "yacht rock",
    "yakousei",
    "yangzhou opera",
    "yaraví",
    "yayue",
    "yé-yé",
    "yodeling",
    "ytpmv",
    "yu-mex",
    "yue opera",
    "yukar",
    "zamacueca",
    "zamba",
    "zamrock",
    "zarzuela",
    "zarzuela barroca",
    "zeitoper",
    "zema",
    "zenonesque",
    "zess",
    "zeuhl",
    "zeybek",
    "zhongguo feng",
    "ziglibithy",
    "zinli",
    "znamenny chant",
    "zoblazo",
    "zohioliin duu",
    "zolo",
    "zouglou",
    "zouk",
    "zouk love",
    "zydeco",
}



"""Module for parsing track names."""

from __future__ import annotations

import operator as op
import re
from collections import Counter
from dataclasses import dataclass, field
from functools import cached_property, reduce
from os.path import commonprefix
from typing import TYPE_CHECKING

from .catalognum import Catalognum
from .helpers import Helpers, cached_patternprop
from .track import Remix

if TYPE_CHECKING:
    from .helpers import JSONDict


@dataclass
class Names:
    """Responsible for parsing track names in the entire release context."""

    # Title [Some Album EP]
    ALBUM_IN_TITLE = cached_patternprop(r"[- ]*\[([^\]]+ [EL]P)\]+", re.I)
    SEPARATOR_PAT = cached_patternprop(r"(?<=\s)[|\u2013\u2014-](?=\s)")
    TITLE_IN_QUOTES = cached_patternprop(r'^(.+[^ -])[ -]+"([^"]+)"$')
    NUMBER_PREFIX = cached_patternprop(r"((?<=^)|(?<=- ))\d{1,2}\W+(?=\D)")

    meta: JSONDict = field(repr=False)
    album_artist: str
    album_in_titles: str | None = None
    catalognum_in_titles: str | None = None
    titles: list[str] = field(default_factory=list)

    @cached_property
    def label(self) -> str:
        try:
            item = self.meta.get("inAlbum", self.meta)["albumRelease"][0]["recordLabel"]
        except (KeyError, IndexError):
            item = self.meta["publisher"]

        return item.get("name") or ""

    @cached_property
    def original_album(self) -> str:
        return str(self.meta["name"])

    @cached_property
    def singleton(self) -> bool:
        return "track" not in self.meta

    @cached_property
    def json_tracks(self) -> list[JSONDict]:
        if self.singleton:
            return [{**self.meta, "byArtist": {"name": self.album_artist}}]

        if tracks := self.meta["track"].get("itemListElement"):
            return [{**t, **t["item"]} for t in tracks]

        # no tracks (sold out release or defunct label, potentially)
        return []

    @cached_property
    def original_titles(self) -> list[str]:
        return [i["name"] for i in self.json_tracks]

    @cached_property
    def _catalognum_in_album_match(self) -> re.Match[str] | None:
        return Catalognum.in_album_pat.search(self.original_album)

    @cached_property
    def catalognum_in_album(self) -> str | None:
        if m := self._catalognum_in_album_match:
            return next(filter(None, m.groups()))

        return None

    @cached_property
    def album(self) -> str:
        if m := self._catalognum_in_album_match:
            return self.original_album.replace(m[0], "")

        return self.original_album

    @cached_property
    def catalognum(self) -> str | None:
        for cat in (self.catalognum_in_album, self.catalognum_in_titles):
            if cat and cat != self.album_artist:
                return cat

        return None

    @property
    def common_prefix(self) -> str:
        return commonprefix(self.titles)

    @classmethod
    def split_quoted_titles(cls, names: list[str]) -> list[str]:
        if len(names) > 1:
            matches = list(filter(None, map(cls.TITLE_IN_QUOTES.match, names)))
            if len(matches) == len(names):
                return [m.expand(r"\1 - \2") for m in matches]

        return names

    def remove_album_catalognum(self, names: list[str]) -> list[str]:
        if catalognum := self.catalognum_in_album:
            pat = re.compile(rf"(?i)[([]{re.escape(catalognum)}[])]")
            return [pat.sub("", n) for n in names]

        return names

    @staticmethod
    def remove_catalognum_with_common_prefix(names: list[str]) -> list[str]:
        matches = list(filter(None, map(Catalognum.in_album_pat.search, names)))
        if len(matches) == len(names) and commonprefix(
            [next(filter(None, m.groups())) for m in matches]
        ):
            cats = [m[0] for m in matches]
            names = [n.replace(c, "").strip("|- ") for n, c in zip(names, cats)]

        return names

    @classmethod
    def remove_number_prefix(cls, names: list[str]) -> list[str]:
        """Remove track number prefix from the track names.

        If there is more than one track and at least half of the track names have
        a number prefix remove it from the names.
        """
        prefix_matches = [cls.NUMBER_PREFIX.search(n) for n in names]
        if len([p for p in prefix_matches if p]) > len(names) / 2:
            return [
                n.replace(p.group() if p else "", "")
                for n, p in zip(names, prefix_matches)
            ]

        return names

    @classmethod
    def find_common_track_delimiter(cls, names: list[str]) -> str:
        """Return the track parts delimiter that is in effect in the current release.

        In some (rare) situations track parts are delimited by a pipe character
        or some UTF-8 equivalent of a dash.

        This checks every track for the first character (see the regex for exclusions)
        that splits it. The character that splits the most and at least half of
        the tracks is the character we need.

        If no such character is found, or if we have just one track, return a dash '-'.
        """

        matches = [m for mat in map(cls.SEPARATOR_PAT.findall, names) for m in mat]
        if not matches:
            return "-"

        delim, count = Counter(matches).most_common(1).pop()
        return delim if (len(names) == 1 or count > len(names) / 2) else "-"

    @classmethod
    def normalize_delimiter(cls, names: list[str]) -> list[str]:
        """Ensure the same delimiter splits artist and title in all names.

        Additionally, assume that a tab character is a delimiter and replace it
        accordingly.
        """
        delim = cls.find_common_track_delimiter(names)
        pat = re.compile(rf"\s+[{re.escape(delim)}]\s+|\t")
        return [pat.sub(" - ", n) for n in names]

    def remove_label(self, names: list[str]) -> list[str]:
        """Remove label name from the end of track names.

        See https://gutterfunkuk.bandcamp.com/album/gutterfunk-all-subject-to-vibes-various-artists-lp
        """
        remove_label = re.compile(rf"([:-]+ |\[){re.escape(self.label)}(\]|$)", re.I)
        return [remove_label.sub(" ", n).strip() for n in names]

    @staticmethod
    def eject_common_catalognum(names: list[str]) -> tuple[str | None, list[str]]:
        """Return catalognum found in every track title.

        1. Split each track name into words
        2. Find the list of words that are common to all tracks
        3. Check the *first* and the *last* word for the catalog number
           - If found, return it and remove it from every track name
        """
        catalognum = None

        names_tokens = [name.split() for name in names]
        sets = [set(tokens) for tokens in names_tokens]
        common_words_set = reduce(op.and_, sets) if sets else set()

        first_name_words = names_tokens[0]
        words = [first_name_words[0], first_name_words[-1]]
        for word in (w for w in words if w in common_words_set):
            if m := Catalognum.anywhere.search(word):
                catalognum = m.group(1)
                names = [n.replace(m.string, "").strip("|- ") for n in names]

        return catalognum, names

    @classmethod
    def eject_album_name(cls, names: list[str]) -> tuple[str | None, list[str]]:
        matches = list(map(cls.ALBUM_IN_TITLE.search, names))
        albums = {m.group(1).replace('"', "") for m in matches if m}
        if len(albums) != 1:
            return None, names

        return albums.pop(), [
            (n.replace(m.group(), "") if m else n) for m, n in zip(matches, names)
        ]

    def ensure_artist_first(self, names: list[str]) -> list[str]:
        """Ensure the artist is the first part of the track name."""
        splits = [n.split(" - ", 1) for n in names]
        left = [s[0] for s in splits]
        if (
            # every track was split at least into two parts
            all(len(s) > 1 for s in splits)
            and (
                (
                    # every track has the same title
                    len(unique_titles := {t for _, t in splits}) == 1
                    # album artists and parts of the unique title overlap
                    and (
                        set(Helpers.split_artists(unique_titles.pop()))
                        & set(Helpers.split_artists(self.album_artist))
                    )
                )
                # or there are at least 2 remixes on the left side of the delimiter
                or sum(bool(Remix.PATTERN.search(x)) for x in left) > 1
            )
        ):
            return [f"{a} - {t}" for t, a in splits]

        return names

    def resolve(self) -> None:
        if not self.original_titles:
            return

        titles = self.split_quoted_titles(self.original_titles)
        if self.singleton:
            titles = [self.album]
        else:
            titles = self.remove_album_catalognum(titles)
            self.catalognum_in_titles, titles = self.eject_common_catalognum(titles)
            titles = self.remove_number_prefix(titles)
            titles = self.remove_catalognum_with_common_prefix(titles)

        titles = self.normalize_delimiter(titles)
        titles = self.remove_label(titles)
        self.album_in_titles, titles = self.eject_album_name(titles)
        self.titles = self.ensure_artist_first(titles)
